from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_preparation.models.constants import (
    BlockLanguage,
    BlockType,
    PipelineType,
)
from mage_ai.data_preparation.templates.utils import (
    read_template_file,
    template_env,
    template_exists,
    write_template,
)
from mage_ai.io.base import DataSource
from typing import Mapping, Union
import json
import os


MAP_DATASOURCE_TO_HANDLER = {
    DataSource.BIGQUERY: 'BigQuery',
    DataSource.POSTGRES: 'Postgres',
    DataSource.REDSHIFT: 'Redshift',
    DataSource.SNOWFLAKE: 'Snowflake',
}


def build_template_from_suggestion(suggestion: Mapping) -> str:
    """
    Creates a file template from a suggestion.

    Args:
        suggestion (Mapping): Suggestion payload generated from `BaseRule`.

    Returns:
        str: String version of Python code to execute to run this cleaning suggestion.
    """
    clean_title = suggestion['title'].lower().replace(' ', '_')
    cleaned_payload = json.dumps(suggestion['action_payload'], indent=4)
    cleaned_payload = cleaned_payload.replace('\n', '\n    ')
    template = read_template_file('transformers/suggestion_fmt.jinja')
    return (
        template.render(
            title=clean_title,
            message=suggestion['message'],
            payload=cleaned_payload,
        )
        + "\n"
    )


def fetch_template_source(
    block_type: Union[BlockType, str],
    config: Mapping[str, str],
    language: BlockLanguage = BlockLanguage.PYTHON,
    pipeline_type: PipelineType = PipelineType.PYTHON,
) -> str:
    template_source = ''

    if BlockLanguage.PYTHON == language:
        if block_type == BlockType.DATA_LOADER:
            template_source = __fetch_data_loader_templates(config, pipeline_type=pipeline_type)
        elif block_type == BlockType.TRANSFORMER:
            template_source = __fetch_transformer_templates(config, pipeline_type=pipeline_type)
        elif block_type == BlockType.DATA_EXPORTER:
            template_source = __fetch_data_exporter_templates(config, pipeline_type=pipeline_type)
        elif block_type == BlockType.SENSOR:
            template_source = __fetch_sensor_templates(config)
    return template_source


def load_template(
    block_type: Union[BlockType, str],
    config: Mapping[str, str],
    dest_path: str,
    language: BlockLanguage = BlockLanguage.PYTHON,
    pipeline_type: PipelineType = PipelineType.PYTHON,
) -> None:
    template_source = fetch_template_source(
        block_type,
        config,
        language=language,
        pipeline_type=pipeline_type,
    )
    write_template(template_source, dest_path)


def __fetch_data_loader_templates(
    config: Mapping[str, str],
    pipeline_type: PipelineType = PipelineType.PYTHON,
) -> str:
    data_source = config.get('data_source')
    if pipeline_type == PipelineType.PYSPARK:
        template_folder = 'data_loaders/pyspark'
    else:
        template_folder = 'data_loaders'

    default_template = os.path.join(template_folder, 'default.jinja')
    if data_source is None:
        template_path = default_template
    else:
        data_source_template = os.path.join(template_folder, f'{data_source.lower()}.py')
        if template_exists(data_source_template):
            template_path = data_source_template
        else:
            template_path = default_template
    return (
        template_env.get_template(template_path).render(
            code=config.get('existing_code', ''),
        )
        + '\n'
    )


def __fetch_transformer_templates(
    config: Mapping[str, str],
    pipeline_type: PipelineType = PipelineType.PYTHON,
) -> str:
    action_type = config.get('action_type')
    axis = config.get('axis')
    data_source = config.get('data_source')
    existing_code = config.get('existing_code', '')
    suggested_action = config.get('suggested_action')

    if suggested_action:
        return build_template_from_suggestion(suggested_action)

    if data_source is not None:
        return __fetch_transformer_data_warehouse_template(data_source)
    elif action_type is not None and axis is not None:
        return __fetch_transformer_action_template(action_type, axis, existing_code)
    else:
        if pipeline_type == PipelineType.PYSPARK:
            template_path = 'transformers/default_pyspark.jinja'
        else:
            template_path = 'transformers/default.jinja'
        return (
            template_env.get_template(template_path).render(
                code=existing_code,
            )
            + '\n'
        )


def __fetch_transformer_data_warehouse_template(data_source: DataSource):
    template = template_env.get_template('transformers/data_warehouse_transformer.jinja')
    data_source_handler = MAP_DATASOURCE_TO_HANDLER.get(data_source)
    if data_source_handler is None:
        raise ValueError(f'No associated database/warehouse for data source \'{data_source}\'')

    if data_source != DataSource.BIGQUERY:
        additional_args = '\n        loader.commit() # Permanently apply database changes'
    else:
        additional_args = ''

    return (
        template.render(
            additional_args=additional_args,
            data_source=data_source if type(data_source) is str else data_source.value,
            data_source_handler=data_source_handler,
        )
        + '\n'
    )


def __fetch_transformer_action_template(action_type: ActionType, axis: Axis, existing_code: str):
    try:
        template = template_env.get_template(
            f'transformers/transformer_actions/{axis}/{action_type}.py'
        )
    except FileNotFoundError:
        template = template_env.get_template('transformers/default.jinja')
    return template.render(code=existing_code) + '\n'


def __fetch_data_exporter_templates(
    config: Mapping[str, str],
    pipeline_type: PipelineType = PipelineType.PYTHON,
) -> str:
    data_source = config.get('data_source')
    if pipeline_type == PipelineType.PYSPARK:
        template_folder = 'data_exporters/pyspark'
    else:
        template_folder = 'data_exporters'

    default_template = os.path.join(template_folder, 'default.jinja')
    if data_source is None:
        template_path = default_template
    else:
        data_source_template = os.path.join(template_folder, f'{data_source.lower()}.py')
        if template_exists(data_source_template):
            template_path = data_source_template
        else:
            template_path = default_template

    return (
        template_env.get_template(template_path).render(
            code=config.get('existing_code', ''),
        )
        + '\n'
    )


def __fetch_sensor_templates(config: Mapping[str, str]) -> str:
    data_source = config.get('data_source')
    try:
        _ = DataSource(data_source)
        template_path = f'sensors/{data_source.lower()}.py'
    except ValueError:
        template_path = 'sensors/default.py'

    return (
        template_env.get_template(template_path).render(
            code=config.get('existing_code', ''),
        )
        + '\n'
    )



from async_timeout import asyncio
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_preparation.models.block import Block, BlockType
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.variable_manager import VariableManager
from mage_ai.tests.base_test import TestCase
from pandas.util.testing import assert_frame_equal
import os
import pandas as pd
import shutil


class BlockTest(TestCase):
    def setUp(self):
        self.repo_path = os.getcwd() + '/test'
        if not os.path.exists(self.repo_path):
            os.mkdir(self.repo_path)
        return super().setUp()

    def tearDown(self):
        shutil.rmtree(self.repo_path)
        return super().tearDown()

    def test_create(self):
        block1 = Block.create('test_transformer', 'transformer', self.repo_path)
        block2 = Block.create('test data loader', BlockType.DATA_LOADER, self.repo_path)
        self.assertTrue(os.path.exists(f'{self.repo_path}/transformers/test_transformer.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/transformers/__init__.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/data_loaders/test_data_loader.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/data_loaders/__init__.py'))
        self.assertEqual(block1.name, 'test_transformer')
        self.assertEqual(block1.uuid, 'test_transformer')
        self.assertEqual(block1.type, 'transformer')
        self.assertEqual(block2.name, 'test data loader')
        self.assertEqual(block2.uuid, 'test_data_loader')
        self.assertEqual(block2.type, 'data_loader')

    def test_execute(self):
        pipeline = Pipeline.create('test pipeline', self.repo_path)
        block1 = Block.create('test_data_loader', 'data_loader', self.repo_path, pipeline=pipeline)
        block2 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 1, 3], 'col2': [2, 2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def remove_duplicate_rows(df):
    df_transformed = df.drop_duplicates()
    return [df_transformed]
            ''')
        asyncio.run(block1.execute())
        asyncio.run(block2.execute())

        variable_manager = VariableManager(pipeline.repo_path)
        data = variable_manager.get_variable(
            pipeline.uuid,
            block2.uuid,
            'df',
            variable_type='dataframe'
        )
        analysis = variable_manager.get_variable(
            pipeline.uuid,
            block2.uuid,
            'df',
            variable_type='dataframe_analysis',
        )
        df_final = pd.DataFrame({'col1': [1, 1, 3], 'col2': [2, 2, 4]}).drop_duplicates()
        assert_frame_equal(data, df_final)
        self.assertEqual(
            analysis['metadata']['column_types'],
            dict(col1=ColumnType.TRUE_OR_FALSE, col2=ColumnType.TRUE_OR_FALSE),
        )
        self.assertTrue(len(analysis['statistics']) > 0)
        self.assertTrue(len(analysis['insights']) > 0)

    def test_execute_multiple_upstream_blocks(self):
        pipeline = Pipeline.create('test pipeline', self.repo_path)
        block1 = Block.create('test_data_loader_1', 'data_loader', self.repo_path, pipeline=pipeline)
        block2 = Block.create('test_data_loader_2', 'data_loader', self.repo_path, pipeline=pipeline)
        block3 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader_1', 'test_data_loader_2'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 3], 'col2': [2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [5], 'col2': [6]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block3.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def union_datasets(df1, df2):
    df_union = pd.concat([df1, df2])
    return [df_union]
            ''')
        asyncio.run(block1.execute())
        asyncio.run(block2.execute())
        asyncio.run(block3.execute())

        variable_manager = VariableManager(pipeline.repo_path)
        data = variable_manager.get_variable(
            pipeline.uuid,
            block3.uuid,
            'df',
            variable_type='dataframe'
        )
        analysis = variable_manager.get_variable(
            pipeline.uuid,
            block3.uuid,
            'df',
            variable_type='dataframe_analysis',
        )
        df_final = pd.concat([
            pd.DataFrame({'col1': [1, 3], 'col2': [2, 4]}),
            pd.DataFrame({'col1': [5], 'col2': [6]}),
        ])
        assert_frame_equal(data, df_final)
        self.assertEqual(
            analysis['metadata']['column_types'],
            dict(col1=ColumnType.NUMBER, col2=ColumnType.NUMBER),
        )
        self.assertTrue(len(analysis['statistics']) > 0)
        self.assertTrue(len(analysis['insights']) > 0)
        self.assertTrue(len(analysis['suggestions']) == 0)

    def test_execute_validation(self):
        pipeline = Pipeline.create('test pipeline', self.repo_path)
        block1 = Block.create('test_data_loader_1', 'data_loader', self.repo_path, pipeline=pipeline)
        block2 = Block.create('test_data_loader_2', 'data_loader', self.repo_path, pipeline=pipeline)
        block3 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader_1', 'test_data_loader_2'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 3], 'col2': [2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [5], 'col2': [6]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block3.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def incorrect_function(df1):
    return df1
            ''')
        asyncio.run(block1.execute())
        asyncio.run(block2.execute())
        with self.assertRaises(Exception):
            asyncio.run(block3.execute())

        with open(block3.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def incorrect_function(df1, df2, df3):
    return df1
            ''')
        with self.assertRaises(Exception):
            asyncio.run(block3.execute())

    def test_to_dict(self):
        block1 = Block.create(
            'test_transformer_2',
            'transformer',
            self.repo_path,
            language='sql',
        )
        block2 = Block.create(
            'test_data_exporter',
            'data_exporter',
            self.repo_path,
            language='python',
        )
        block2.upstream_blocks = [block1]
        block1.downstream_blocks = [block2]
        self.assertEqual(block1.to_dict(), dict(
            all_upstream_blocks_executed=True,
            configuration={},
            downstream_blocks=['test_data_exporter'],
            executor_config=None,
            executor_type='local_python',
            language='sql',
            name='test_transformer_2',
            status='not_executed',
            type='transformer',
            upstream_blocks=[],
            uuid='test_transformer_2',
        ))
        self.assertEqual(block2.to_dict(), dict(
            all_upstream_blocks_executed=False,
            configuration={},
            downstream_blocks=[],
            executor_config=None,
            executor_type='local_python',
            language='python',
            name='test_data_exporter',
            status='not_executed',
            type='data_exporter',
            upstream_blocks=['test_transformer_2'],
            uuid='test_data_exporter',
        ))



from enum import Enum


class DataType(str, Enum):
    DATA_FRAME = 'data_frame'
    IMAGE_PNG = 'image/png'
    PROGRESS = 'progress'
    TABLE = 'table'
    TEXT = 'text'
    TEXT_HTML = 'text/html'
    TEXT_PLAIN = 'text/plain'

COMMS_MESSAGE_TYPES = [
    'comm_open',
    'comm_msg',
    'comm_close',
]


def parse_output_message(message: dict) -> dict:
    data_content = None
    data_type = None
    error = None

    header = message['header']
    msg_type = header['msg_type']

    parent_header = message['parent_header']
    msg_id = parent_header.get('msg_id')

    content = message['content']
    execution_state = content.get('execution_state')

    traceback = content.get('traceback')
    data = content.get('data', {})
    metadata = content.get('metadata')
    text = data.get('text/plain')
    text_html = data.get('text/html')
    code = data.get('code')
    image = data.get('image/png')

    if content.get('name') in ['stdout', 'stderr']:
        text_stdout = content.get('text')
        data_content = text_stdout.split('\n')
        data_type = DataType.TEXT_PLAIN
    elif image:
        data_content = image
        data_type = DataType.IMAGE_PNG
    elif traceback:
        data_content = [line for line in traceback]
        data_type = DataType.TEXT
        error = traceback
    elif text_html:
        data_content = text_html
        data_type = DataType.TEXT_HTML
    elif text:
        data_content = text.split('\n')
        if len(data_content[-1]) != 0:
            data_content.append('')
        data_type = DataType.TEXT_PLAIN
    elif code:
        data_content = code
        data_type = DataType.TEXT
    elif msg_type in COMMS_MESSAGE_TYPES:
        if data.get('method') == 'update':
            progress_data = f"{data.get('state', dict()).get('value', 0) * 100}"
            data_content = progress_data
            data_type = DataType.PROGRESS

    return dict(
        data=data_content,
        error=error,
        execution_state=execution_state,
        metadata=metadata,
        msg_id=msg_id,
        msg_type=msg_type,
        type=data_type,
    )



from .constants import TimeInterval, TIME_INTERVAL_TO_TIME_DELTA
from .utils import calculate_metric_for_series, clean_series
from datetime import datetime
from dateutil.relativedelta import relativedelta
import dateutil.parser
import math
import numpy as np
import pandas as pd


MAX_BUCKETS = 40
TIME_SERIES_BUCKETS = 40


def build_buckets(min_value, max_value, max_buckets):
    diff = max_value - min_value
    total_interval = 1 + diff
    bucket_interval = total_interval / max_buckets
    number_of_buckets = max_buckets

    is_integer = False
    parts = str(diff).split('.')
    if len(parts) == 1:
        is_integer = True
    else:
        is_integer = int(parts[1]) == 0

    if is_integer and total_interval <= max_buckets:
        number_of_buckets = int(total_interval)
        bucket_interval = 1
    elif bucket_interval > 1:
        bucket_interval = math.ceil(bucket_interval)
    else:
        bucket_interval = round(bucket_interval * 100, 1) / 100

    buckets = []
    for i in range(number_of_buckets):
        min_v = min_value + (i * bucket_interval)
        max_v = min_value + ((i + 1) * bucket_interval)
        if max_value >= min_v:
            buckets.append(dict(
                max_value=max_v,
                min_value=min_v,
                values=[],
            ))

    return buckets, bucket_interval


def build_histogram_data(arr, max_buckets):
    max_value = max(arr)
    min_value = min(arr)

    buckets, bucket_interval = build_buckets(min_value, max_value, max_buckets)

    if bucket_interval == 0:
        return

    bins = [b['min_value'] for b in buckets] + [buckets[-1]['max_value']]
    count, _ = np.histogram(arr, bins=bins)

    x = []
    y = []

    for idx, bucket in enumerate(buckets):
        x.append(dict(
            max=bucket['max_value'],
            min=bucket['min_value'],
        ))
        y.append(dict(value=count[idx]))

    return dict(
        x=x,
        y=y,
    )


def build_time_series_buckets(df, datetime_column, time_interval, metrics):
    time_values = df[datetime_column]
    datetimes = clean_series(time_values)
    if datetimes.size <= 1:
        return []

    datetimes = datetimes.unique()
    min_value_datetime = dateutil.parser.parse(datetimes.min())
    max_value_datetime = dateutil.parser.parse(datetimes.max())

    a, b = [dateutil.parser.parse(d) for d in sorted(datetimes)[:2]]

    year = min_value_datetime.year
    month = min_value_datetime.month
    day = min_value_datetime.day
    hour = min_value_datetime.hour
    minute = min_value_datetime.minute

    start_datetime = min_value_datetime

    if TimeInterval.ORIGINAL == time_interval:
        diff = (b - a).total_seconds()
        if diff >= 60 * 60 * 24 * 365:
            time_interval = TimeInterval.YEAR
        elif diff >= 60 * 60 * 24 * 30:
            time_interval = TimeInterval.MONTH
        elif diff >= 60 * 60 * 24 * 7:
            time_interval = TimeInterval.WEEK
        elif diff >= 60 * 60 * 24:
            time_interval = TimeInterval.DAY
        elif diff >= 60 * 60:
            time_interval = TimeInterval.HOUR
        elif diff >= 60:
            time_interval = TimeInterval.SECOND
        else:
            time_interval = TimeInterval.SECOND

    if TimeInterval.DAY == time_interval:
        start_datetime = datetime(year, month, day, 0, 0, 0)
    elif TimeInterval.HOUR == time_interval:
        start_datetime = datetime(year, month, day, hour, 0, 0)
    elif TimeInterval.MINUTE == time_interval:
        start_datetime = datetime(year, month, day, hour, minute, 0)
    elif TimeInterval.MONTH == time_interval:
        start_datetime = datetime(year, month, 1, 0, 0, 0)
    elif TimeInterval.SECOND == time_interval:
        start_datetime = datetime(year, month, day, hour, minute, 0)
    elif TimeInterval.WEEK == time_interval:
        start_datetime = min_value_datetime - relativedelta(
            days=min_value_datetime.isocalendar()[2],
        )
        start_datetime = datetime(
            start_datetime.year,
            start_datetime.month,
            start_datetime.day,
            0,
            0,
            0,
        )
    elif TimeInterval.YEAR == time_interval:
        start_datetime = datetime(year, 1, 1, 0, 0, 0)

    df_copy = df.copy()
    df_copy[datetime_column] = \
        pd.to_datetime(df[datetime_column]).apply(lambda x: x if pd.isnull(x) else x.timestamp())

    values = [[] for _ in metrics]
    buckets = []

    while len(buckets) == 0 or buckets[-1] <= max_value_datetime.timestamp():
        if len(buckets) == 0:
            min_date_ts = start_datetime.timestamp()
        else:
            min_date_ts = buckets[-1]
        max_date = datetime.fromtimestamp(min_date_ts) + TIME_INTERVAL_TO_TIME_DELTA[time_interval]
        buckets.append(max_date.timestamp())

        df_in_range = df_copy[(
            df_copy[datetime_column] >= min_date_ts
        ) & (
            df_copy[datetime_column] < max_date.timestamp()
        )]

        for idx, metric in enumerate(metrics):
            aggregation = metric['aggregation']
            column = metric['column']
            series = df_in_range[column]
            values[idx].append(calculate_metric_for_series(series, aggregation))

    return buckets, values



from .base import BaseHandler
from mage_ai.shared.hash import merge_dict


class ApiClustersHandler(BaseHandler):
    def get(self, cluster_type):
        clusters = []
        if cluster_type == 'emr':
            from mage_ai.cluster_manager.aws.emr_cluster_manager import emr_cluster_manager
            clusters = emr_cluster_manager.list_clusters()
        self.write(dict(clusters=clusters))

    def post(self, cluster_type):
        success = False

        if cluster_type == 'emr':
            from mage_ai.cluster_manager.aws.emr_cluster_manager import emr_cluster_manager
            cluster_payload = self.get_payload().get('cluster')
            if cluster_payload is None:
                raise Exception('Please include cluster info in the request payload')
            action = cluster_payload.get('action')
            if action == 'create_new_cluster':
                cluster_id = emr_cluster_manager.create_cluster()['cluster_id']
                success = True
            elif action == 'set_active_cluster':
                cluster_id = cluster_payload.get('cluster_id')
                if cluster_id is None:
                    raise Exception('Please include cluster_id in thhe request payhload')
                emr_cluster_manager.set_active_cluster(cluster_id)
                success = True

        self.write(dict(
            cluster=merge_dict(dict(id=cluster_id), cluster_payload),
            success=success,
        ))

    def put(self, cluster_type):
        if cluster_type == 'emr':
            from mage_ai.cluster_manager.aws.emr_cluster_manager import emr_cluster_manager
            cluster_payload = self.get_payload().get('cluster')
            if cluster_payload is None:
                raise Exception('Please include cluster info in the request payload')
            action = cluster_payload.get('action')

            cluster_id = cluster_payload.get('id')
            if cluster_id is None:
                raise Exception('Please include cluster_id in thhe request payhload')
            emr_cluster_manager.set_active_cluster(cluster_id)
            success = True

        self.write(dict(
            cluster=merge_dict(dict(id=cluster_id), cluster_payload),
            success=success,
        ))



from mage_ai.orchestration.run_status_checker import check_status

if 'sensor' not in globals():
    from mage_ai.data_preparation.decorators import sensor


@sensor
def check_condition(**kwargs) -> bool:
    """
    Template code for checking if block or pipeline run completed.
    """
    return check_status(
        'pipeline_uuid',
        kwargs['execution_date'],
        block_uuid='block_uuid',  # optional
    )



from mage_ai.shared.hash import merge_dict

import json
import logging
import requests

logger = logging.getLogger(__name__)


class Mage:
    def __init__(self, **kwargs):
        self.url_prefix = 'https://backend.mage.ai/api/v1'

    def sync_pipeline(self, pipeline, api_key):
        error_message = f'Syncing pipeline {pipeline.id} failed'
        if api_key is None:
            logger.error(f'{error_message}, invalid API key')
            return
        feature_set = pipeline.get_feature_set()
        if feature_set is None:
            logger.error(f'{error_message}, feature set does not exist')
            return
        pipeline_name = f"{feature_set.metadata['name']}_pipeline"
        data = {
            'data_cleaning_pipeline': {
                'name': pipeline_name,
                'pipeline_actions': pipeline.pipeline.actions,
            }
        }
        try:
            remote_id = pipeline.metadata.get('remote_id')
            if remote_id is not None:
                requests.put(
                    data=json.dumps(data),
                    headers={
                        'Content-Type': 'application/json',
                        'X-API-KEY': api_key,
                    },
                    url=f'{self.url_prefix}/data_cleaning_pipelines/{remote_id}',
                )
            else:
                response = requests.post(
                    data=json.dumps(data),
                    headers={
                        'Content-Type': 'application/json',
                        'X-API-KEY': api_key,
                    },
                    url=f'{self.url_prefix}/data_cleaning_pipelines',
                ).json()
                pipeline_response = response['data_cleaning_pipeline']
                pipeline.metadata = merge_dict(
                    pipeline.metadata,
                    {
                        'remote_id': pipeline_response['id'],
                    },
                )
        except:
            logger.exception(error_message)
            pass

    def get_pipeline_actions(self, id, api_key):
        if api_key is None:
            logger.error('Fetching pipeline actions failed, invalid API key')
            return []
        try:
            response = requests.get(
                headers={
                    'Content-Type': 'application/json',
                    'X-API-KEY': api_key,
                },
                url=f'{self.url_prefix}/data_cleaning_pipelines/{id}',
            ).json()
            return response['data_cleaning_pipeline'].get('pipeline_actions', [])
        except:
            logger.exception('Fetching pipeline actions from database failed')
            return []



from enum import Enum

PIPELINES_FOLDER = 'pipelines'
PIPELINE_CONFIG_FILE = 'metadata.yaml'


DATAFRAME_ANALYSIS_KEYS = frozenset(
    [
        'metadata',
        'statistics',
        'insights',
        'suggestions',
    ]
)
DATAFRAME_ANALYSIS_MAX_ROWS = 100_000
DATAFRAME_ANALYSIS_MAX_COLUMNS = 30
DATAFRAME_SAMPLE_COUNT_PREVIEW = 10
DATAFRAME_SAMPLE_COUNT = 1000
VARIABLE_DIR = '.variables'
LOGS_DIR = '.logs'


class BlockLanguage(str, Enum):
    PYTHON = 'python'
    SQL = 'sql'


class BlockStatus(str, Enum):
    EXECUTED = 'executed'
    FAILED = 'failed'
    NOT_EXECUTED = 'not_executed'
    UPDATED = 'updated'


class BlockType(str, Enum):
    CHART = 'chart'
    DATA_EXPORTER = 'data_exporter'
    DATA_LOADER = 'data_loader'
    SCRATCHPAD = 'scratchpad'
    SENSOR = 'sensor'
    TRANSFORMER = 'transformer'


class ExecutorType(str, Enum):
    LOCAL_PYTHON = 'local_python'
    ECS = 'ecs'
    K8S = 'k8s'
    PYSPARK = 'pyspark'


class PipelineType(str, Enum):
    PYTHON = 'python'
    PYSPARK = 'pyspark'


CUSTOM_EXECUTION_BLOCK_TYPES = [
    BlockType.CHART,
    BlockType.DATA_EXPORTER,
    BlockType.DATA_LOADER,
    BlockType.SENSOR,
    BlockType.TRANSFORMER,
]


NON_PIPELINE_EXECUTABLE_BLOCK_TYPES = [
    BlockType.CHART,
    BlockType.SCRATCHPAD,
]



TEST_ACTION = dict(
    action_type='filter',
    axis='row',
    action_code='%{1}.%{1_1} == True and (%{1}.%{1_2} == \"The Quant\" or %{1}.%{1_2} == \"Yield\")',
    action_arguments=[
        '%{1}.%{1_1}',
        '%{2}.%{2_1}',
    ],
    action_options=dict(
        condition='%{1}.%{1_3} >= %{2}.%{2_2} and %{2}.%{2_2} >= %{1}.%{1_3} - 2592000',
        default=0,
        timestamp_feature_a='%{1}.%{1_2}',
        timestamp_feature_b='%{1}.%{1_3}',
        window=2592000,
    ),
    action_variables={
        '1': dict(type='feature_set_version', feature_set_version=dict(
            feature_set=dict(column_type='category', uuid='omni'),
        )),
        '1_1': dict(type='feature', feature=dict(column_type='number', uuid='deposited')),
        '1_2': dict(type='feature', feature=dict(column_type='category', uuid='fund')),
        '1_3': dict(type='feature', feature=dict(column_type='category', uuid='delivered_at')),
        '1_4': dict(type='feature', feature=dict(column_type='number_with_decimals', uuid='amount')),
        '2': dict(type='feature_set_version', feature_set_version=dict(
            feature_set=dict(column_type='category', uuid='magic'),
        )),
        '2_1': dict(type='feature', feature=dict(column_type='category', uuid='spell')),
        '2_2': dict(type='feature', feature=dict(column_type='category', uuid='booked_at')),
        '3_1': dict(type='feature', feature=dict(column_type='number', uuid='age')),
    },
)



from mage_ai.data_cleaner.shared.utils import (
    is_geo_dataframe,
    is_spark_dataframe,
)
from mage_ai.data_preparation.models.variable import Variable, VariableType, VARIABLE_DIR
from mage_ai.data_preparation.repo_manager import get_repo_path
from typing import Any, Dict, List
import os
import pandas as pd


class VariableManager:
    def __init__(self, repo_path=None, variables_dir=None):
        self.repo_path = repo_path or get_repo_path()
        if variables_dir is None:
            self.variables_dir = self.repo_path
        else:
            self.variables_dir = variables_dir
        # TODO: implement caching logic

    def add_variable(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        data: Any,
        partition: str = None,
        variable_type: VariableType = None
    ) -> None:
        if type(data) is pd.DataFrame:
            variable_type = VariableType.DATAFRAME
        elif is_spark_dataframe(data):
            variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            variable_type = VariableType.GEO_DATAFRAME
        variable = Variable(
            variable_uuid,
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            variable_type=variable_type,
        )
        # Delete data if it exists
        variable.delete()
        variable.variable_type = variable_type
        variable.write_data(data)

    def delete_variable(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        partition: str = None,
        variable_type: VariableType = None,
    ) -> None:
        Variable(
            variable_uuid,
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            variable_type=variable_type,
        ).delete()

    def get_variable(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        partition: str = None,
        variable_type: VariableType = None,
        sample: bool = False,
        sample_count: int = None,
        spark=None,
    ) -> Any:
        variable = self.get_variable_object(
            pipeline_uuid,
            block_uuid,
            variable_uuid,
            partition=partition,
            variable_type=variable_type,
            spark=spark,
        )
        return variable.read_data(sample=sample, sample_count=sample_count, spark=spark)

    def get_variable_object(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        variable_uuid: str,
        partition: str = None,
        variable_type: VariableType = None,
        spark=None,
    ) -> Variable:
        if variable_type == VariableType.DATAFRAME and spark is not None:
            variable_type = VariableType.SPARK_DATAFRAME
        return Variable(
            variable_uuid,
            self.__pipeline_path(pipeline_uuid),
            block_uuid,
            partition=partition,
            variable_type=variable_type,
        )

    def get_variables_by_pipeline(self, pipeline_uuid: str) -> Dict[str, List[str]]:
        from mage_ai.data_preparation.models.pipeline import Pipeline
        pipeline = Pipeline.get(pipeline_uuid, repo_path=self.repo_path)
        variable_dir_path = os.path.join(self.__pipeline_path(pipeline_uuid), VARIABLE_DIR)
        if not os.path.exists(variable_dir_path):
            return dict()
        block_dirs = os.listdir(variable_dir_path)
        variables_by_block = dict()
        for d in block_dirs:
            if not pipeline.has_block(d) and d != 'global':
                continue
            block_variables_path = os.path.join(variable_dir_path, d)
            if not os.path.isdir(block_variables_path):
                variables_by_block[d] = []
            else:
                variables = os.listdir(os.path.join(variable_dir_path, d))
                variable_names = sorted([v.split('.')[0] for v in variables])
                variables_by_block[d] = [v for v in variable_names if v != '']
        return variables_by_block

    def get_variables_by_block(
        self,
        pipeline_uuid: str,
        block_uuid: str,
        partition: str = None,
    ) -> List[str]:
        variable_dir_path = os.path.join(
            self.__pipeline_path(pipeline_uuid),
            VARIABLE_DIR,
            partition or '',
            block_uuid,
        )
        if not os.path.exists(variable_dir_path):
            return []
        variables = os.listdir(variable_dir_path)
        return sorted([v.split('.')[0] for v in variables])

    def __pipeline_path(self, pipeline_uuid: str) -> str:
        path = os.path.join(self.variables_dir, 'pipelines', pipeline_uuid)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        return path


def get_global_variables(
    pipeline_uuid: str,
    repo_path: str = None,
) -> Dict[str, Any]:
    if repo_path is None:
        repo_path = get_repo_path()
    variables = VariableManager(repo_path).get_variables_by_block(pipeline_uuid, 'global')
    global_variables = dict()
    for variable in variables:
        global_variables[variable] = get_global_variable(pipeline_uuid, variable, repo_path)

    return global_variables


def get_global_variable(
    pipeline_uuid: str,
    key: str,
    repo_path: str = None,
) -> Any:
    if repo_path is None:
        repo_path = get_repo_path()
    return VariableManager(repo_path).get_variable(pipeline_uuid, 'global', key)


def get_variable(
    pipeline_uuid: str,
    block_uuid: str,
    key: str,
    **kwargs
) -> Any:
    return VariableManager(get_repo_path()).get_variable(
        pipeline_uuid,
        block_uuid,
        key,
        **kwargs,
    )


def set_global_variable(
    pipeline_uuid: str,
    key: str,
    value: Any,
    repo_path: str = None,
) -> None:
    if repo_path is None:
        repo_path = get_repo_path()
    VariableManager(repo_path).add_variable(pipeline_uuid, 'global', key, value)


def delete_global_variable(
    pipeline_uuid: str,
    key: str,
    repo_path: str = None,
) -> None:
    if repo_path is None:
        repo_path = get_repo_path()
    VariableManager(repo_path).delete_variable(pipeline_uuid, 'global', key)



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.snowflake import Snowflake
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_snowflake(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to a Snowflake warehouse.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#snowflake
    """
    table_name = 'your_table_name'
    database = 'your_database_name'
    schema = 'your_schema_name'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    with Snowflake.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        loader.export(
            df,
            table_name,
            database,
            schema,
            if_exists='replace',  # Specify resolution policy if table already exists
        )



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.s3 import S3
from os import path

import time


if 'sensor' not in globals():
    from mage_ai.data_preparation.decorators import sensor


@sensor
def check_condition(**kwargs) -> bool:
    """
    Template code for checking if a partition exists in a S3 bucket
    """

    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    bucket_name = 'your_bucket_name'
    path = 'your_partition_path'

    return S3.with_config(ConfigFileLoader(config_path, config_profile)).exists(
        bucket_name, path
    )



from abc import ABC, abstractmethod
from mage_ai.orchestration.pipeline_scheduler import schedule_all


class TimeTrigger(ABC):
    def __init__(self, trigger_interval: int = 10) -> None:
        self.trigger_interval = trigger_interval

    def run(self) -> None:
        schedule_all()

    @abstractmethod
    def start(self) -> None:
        pass



from mage_ai.orchestration.db.models import PipelineRun
import multiprocessing


class ExecutionProcessManager:
    def __init__(self):
        self.processes = dict()

    def set_process(self, pipeline_run_id: int, block_run_id: int, proc: multiprocessing.Process):
        if pipeline_run_id not in self.processes:
            self.processes[pipeline_run_id] = dict()
        self.processes[pipeline_run_id][block_run_id] = proc

    def clean_up_processes(self):
        """
        Clean up inactive processes and terminate processes for cancelled pipeline runs
        """
        for pipeline_run_id in list(self.processes.keys()):
            block_run_procs = self.processes[pipeline_run_id]
            if not block_run_procs:
                del self.processes[pipeline_run_id]
                continue
            pipeline_run = PipelineRun.query.get(pipeline_run_id)
            if pipeline_run.status == PipelineRun.PipelineRunStatus.CANCELLED:
                for block_run_id in list(block_run_procs.keys()):
                    proc = block_run_procs[block_run_id]
                    if proc.is_alive():
                        proc.terminate()
                    del block_run_procs[block_run_id]
            else:
                for block_run_id in list(block_run_procs.keys()):
                    proc = block_run_procs[block_run_id]
                    if not proc.is_alive():
                        del block_run_procs[block_run_id]
            if not block_run_procs:
                del self.processes[pipeline_run_id]


execution_process_manager = ExecutionProcessManager()



from mage_ai.data_cleaner.cleaning_rules.impute_values import ImputeValues
from mage_ai.data_cleaner.shared.utils import clean_dataframe
from mage_ai.tests.base_test import TestCase
import numpy as np
import pandas as pd


class ImputeValuesTest(TestCase):
    def setUp(self):
        self.rng = np.random.default_rng(42)
        return super().setUp()

    def test_mean(self):
        df = pd.DataFrame(
            [
                [10, 3],
                [None, None],
                [21, None],
                [25, None],
                [None, None],
                [np.NaN, None],
                [31, None],
                [25, 9],
                [21, 99],
                [10, 2],
            ],
            columns=['age', 'number_of_years'],
        )
        column_types = {'age': 'number', 'number_of_years': 'number'}
        statistics = {
            'age/count': 7,
            'age/count_distinct': 4,
            'age/null_value_rate': 3 / 10,
            'number_of_years/count': 4,
            'number_of_years/count_distinct': 4,
            'number_of_years/null_value_rate': 6 / 10,
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='For each column, fill missing entries with the average value.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['age'],
                    action_options=dict(strategy='average'),
                    action_variables=dict(
                        age=dict(feature=dict(column_type='number', uuid='age'), type='feature')
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Fill in missing values',
                message='Fill missing values with a placeholder to mark them as missing.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['number_of_years'],
                    action_options=dict(strategy='constant'),
                    action_variables=dict(
                        number_of_years=dict(
                            feature=dict(column_type='number', uuid='number_of_years'),
                            type='feature',
                        )
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_mean_large(self):
        age_arr = self.rng.integers(0, 110, size=(150,)).astype(float)
        indices = np.arange(0, 150)
        self.rng.shuffle(indices)
        age_arr[indices[:49]] = np.NaN

        num_years_arr = self.rng.integers(0, 110, size=(150,)).astype(float)
        self.rng.shuffle(indices)
        num_years_arr[indices[:50]] = np.NaN
        df = pd.DataFrame({'age': age_arr, 'number_of_years': num_years_arr})
        column_types = {'age': 'number', 'number_of_years': 'number'}
        statistics = {
            'age/count': df['age'].count(),
            'age/count_distinct': df['age'].nunique(),
            'age/null_value_rate': 1 - df['age'].count() / len(df['age']),
            'number_of_years/count': df['number_of_years'].count(),
            'number_of_years/count_distinct': df['number_of_years'].nunique(),
            'number_of_years/null_value_rate': (
                1 - df['number_of_years'].count() / len(df['number_of_years'])
            ),
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='For each column, fill missing entries with the average value.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['age'],
                    action_options=dict(strategy='average'),
                    action_variables=dict(
                        age=dict(feature=dict(column_type='number', uuid='age'), type='feature')
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Fill in missing values',
                message='Fill missing values with a placeholder to mark them as missing.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['number_of_years'],
                    action_options=dict(strategy='constant'),
                    action_variables=dict(
                        number_of_years=dict(
                            feature=dict(column_type='number', uuid='number_of_years'),
                            type='feature',
                        )
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_median(self):
        df = pd.DataFrame(
            [
                [90.5, 3],
                [None, None],
                [52.3, None],
                [-2.3, None],
                [None, None],
                [np.NaN, None],
                [-1.4, None],
                [-1.6, 9],
                [-4.3, 9],
                [-3.4, 9],
            ],
            columns=['profit', 'number_of_years'],
        )
        column_types = {'profit': 'number_with_decimals', 'number_of_years': 'number'}
        statistics = {
            'profit/count': 7,
            'profit/count_distinct': 7,
            'profit/null_value_rate': 0.3,
            'number_of_years/count': 2,
            'number_of_years/count_distinct': 2,
            'number_of_years/null_value_rate': 0.6,
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='Fill missing values with a placeholder to mark them as missing.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['number_of_years'],
                    action_options=dict(strategy='constant'),
                    action_variables=dict(
                        number_of_years=dict(
                            feature=dict(column_type='number', uuid='number_of_years'),
                            type='feature',
                        )
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Fill in missing values',
                message='For each column, fill missing entries with the median value.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['profit'],
                    action_options=dict(strategy='median'),
                    action_variables=dict(
                        profit=dict(
                            feature=dict(column_type='number_with_decimals', uuid='profit'),
                            type='feature',
                        )
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_mode(self):
        df = pd.DataFrame(
            [
                [90.5, 'one', 34934],
                [np.nan, None, np.nan],
                [667, 'one', 10010],
                [np.nan, 'one', 34934],
                [-234, None, np.nan],
                [np.nan, None, 34934],
                [-1.4, 'one', np.nan],
                [-1.6, 'None', 10010],
            ],
            columns=['profit', 'company', 'industry'],
        )
        column_types = {
            'profit': 'number_with_decimals',
            'company': 'category',
            'industry': 'zip_code',
        }
        statistics = {
            'profit/count': 5,
            'profit/null_value_rate': 3 / 8,
            'company/count': 5,
            'company/null_value_rate': 3 / 8,
            'company/mode': 'one',
            'company/mode_ratio': 4 / 5,
            'industry/count': 5,
            'industry/null_value_rate': 3 / 8,
            'industry/mode': 34934,
            'industry/mode_ratio': 3 / 5,
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='Fill missing values with a placeholder to mark them as missing.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['profit'],
                    action_options=dict(strategy='constant'),
                    action_variables=dict(
                        profit=dict(
                            feature=dict(column_type='number_with_decimals', uuid='profit'),
                            type='feature',
                        )
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Fill in missing values',
                message='For each column, fill missing entries with the most frequent value.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['company', 'industry'],
                    action_options=dict(strategy='mode'),
                    action_variables=dict(
                        company=dict(
                            feature=dict(column_type='category', uuid='company'), type='feature'
                        ),
                        industry=dict(
                            feature=dict(column_type='zip_code', uuid='industry'), type='feature'
                        ),
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_no_numerical(self):
        df = pd.DataFrame(
            [
                [90.5, 3],
                [None, ''],
                ['', None],
                [np.NaN, None],
                [None, None],
                [np.NaN, None],
                [-1.4, None],
                [-1.6, 9],
            ],
            columns=['profit', 'number_of_years'],
        )
        column_types = {'profit': 'number_with_decimals', 'number_of_years': 'number'}
        statistics = {
            'profit/count': 3,
            'profit/count_distinct': 3,
            'profit/null_value_rate': 5 / 8,
            'number_of_years/count': 2,
            'number_of_years/count_distinct': 2,
            'number_of_years/null_value_rate': 6 / 8,
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='Fill missing values with a placeholder to mark them as missing.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['profit', 'number_of_years'],
                    action_options=dict(strategy='constant'),
                    action_variables=dict(
                        profit=dict(
                            feature=dict(column_type='number_with_decimals', uuid='profit'),
                            type='feature',
                        ),
                        number_of_years=dict(
                            feature=dict(column_type='number', uuid='number_of_years'),
                            type='feature',
                        ),
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_random(self):
        source_ids = np.concatenate(
            [self.rng.integers(10000, 99999, size=(20,)).astype(str) for _ in range(10)]
        )
        source_idxs = self.rng.integers(0, 100, size=(60,))
        source_ids[source_idxs] = ''
        dest_ids = np.concatenate(
            [self.rng.integers(10000, 99999, size=(20,)).astype(str) for _ in range(10)]
        )
        dest_idxs_seq = np.arange(0, 199, 8)
        dest_idxs_rand = self.rng.integers(0, 99, size=(70))
        dest_ids[dest_idxs_seq] = ''
        dest_ids[dest_idxs_rand] = ''
        df = pd.DataFrame({'source': source_ids, 'dest': dest_ids})
        cleaned_df = df.applymap(
            lambda x: x if (not isinstance(x, str) or (len(x) > 0 and not x.isspace())) else np.nan
        )

        column_types = {'source': 'category_high_cardinality', 'dest': 'category_high_cardinality'}
        statistics = {
            'source/count': cleaned_df['source'].count(),
            'source/count_distinct': len(cleaned_df['source'].value_counts().index) - 1,
            'source/null_value_rate': 1 - cleaned_df['source'].count() / len(df),
            'source/mode': cleaned_df['source'].mode(),
            'source/mode_ratio': cleaned_df['source'].value_counts().max()
            / cleaned_df['source'].count(),
            'dest/count': cleaned_df['dest'].count(),
            'dest/count_distinct': len(cleaned_df['dest'].value_counts().index) - 1,
            'dest/null_value_rate': 1 - cleaned_df['dest'].count() / len(df),
            'dest/mode': cleaned_df['dest'].mode(),
            'dest/mode_ratio': cleaned_df['dest'].value_counts().max() / cleaned_df['dest'].count(),
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='Fill missing values with a placeholder to mark them as missing.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['dest'],
                    action_options=dict(strategy='constant'),
                    action_variables=dict(
                        dest=dict(
                            feature=dict(column_type='category_high_cardinality', uuid='dest'),
                            type='feature',
                        ),
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
            dict(
                title='Fill in missing values',
                message='For each column, fill missing entries with randomly sampled values.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['source'],
                    action_options=dict(strategy='random'),
                    action_variables=dict(
                        source=dict(
                            feature=dict(column_type='category_high_cardinality', uuid='source'),
                            type='feature',
                        )
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            ),
        ]
        df = clean_dataframe(df, column_types, dropna=False)
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_row_rm(self):
        df = pd.DataFrame(
            [
                ['CT', '06902'],
                ['NY', '10001'],
                ['MI', '48841'],
                ['CA', '95001'],
                ['', None],
                [None, '23456'],
                ['CA', None],
                ['MA', '12214'],
                ['PA', '37821'],
                ['TX', '75001'],
            ],
            columns=['state', 'location'],
        )
        column_types = {'state': 'category', 'location': 'zip_code'}
        statistics = {
            'state/count': 8,
            'state/count_distinct': 7,
            'state/null_value_rate': 0.2,
            'location/count': 8,
            'location/count_distinct': 8,
            'location/null_value_rate': 0.2,
            'is_timeseries': False,
        }
        expected_suggestions = [
            dict(
                title='Remove rows with missing entries',
                message='Delete 3 rows to remove all missing values from the dataset.',
                action_payload=dict(
                    action_type='filter',
                    action_arguments=['state', 'location'],
                    action_options={},
                    action_variables=dict(
                        state=dict(
                            feature=dict(column_type='category', uuid='state'), type='feature'
                        ),
                        location=dict(
                            feature=dict(column_type='zip_code', uuid='location'), type='feature'
                        ),
                    ),
                    action_code='state != null and location != null',
                    axis='row',
                    outputs=[],
                ),
                status='not_applied',
            )
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_seq(self):
        df = pd.DataFrame(
            [
                ['CT', '06902', '12-24-2022'],
                ['NY', '10001', '12-25-2022'],
                [None, '', None],
                ['CA', '', '12-28-2022'],
                [None, '', '12-30-2022'],
                ['CA', None, '12-30-2022'],
                ['MA', '12214', '12-31-2022'],
                ['PA', '', '1-2-2023'],
                ['TX', '75001', '1-2-2023'],
                ['', None, ''],
            ],
            columns=['state', 'location', 'timestamp'],
        )
        df['lists'] = pd.Series(
            [
                (np.nan, 2.0, 'string', '3'),
                ['not string?', True, True, 8, False, np.nan, None],
                None,
                None,
                None,
                ('not string?', True, True, 8, False, np.nan, None),
                [],
                '[\'not string?\'   ,  True, True , 8   , False  , np.nan, None]',
                '(\'not string?\'   ,  True, True , 8   , False  , np.nan, None)',
                tuple(),
            ]
        )
        column_types = {
            'state': 'category',
            'location': 'zip_code',
            'timestamp': 'datetime',
            'lists': 'list',
        }
        statistics = {
            'state/count': 7,
            'state/count_distinct': 6,
            'state/null_value_rate': 0.4,
            'state/max_null_seq': 1,
            'location/count': 4,
            'location/count_distinct': 4,
            'location/null_value_rate': 0.6,
            'location/max_null_seq': 4,
            'timestamp/null_value_rate': 1 / 10,
            'timestamp/max_null_seq': 1,
            'lists/null_value_rate': 0.3,
            'lists/max_null_seq': 1,
            'lists/mode_ratio': 2 / 10,
            'is_timeseries': True,
            'timeseries_index': ['timestamp'],
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='Fill missing entries using the previously occurring '
                'entry in the timeseries.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['state', 'location', 'timestamp', 'lists'],
                    action_options=dict(strategy='sequential', timeseries_index=['timestamp']),
                    action_variables=dict(
                        state=dict(
                            feature=dict(column_type='category', uuid='state'), type='feature'
                        ),
                        location=dict(
                            feature=dict(column_type='zip_code', uuid='location'), type='feature'
                        ),
                        timestamp=dict(
                            feature=dict(column_type='datetime', uuid='timestamp'), type='feature'
                        ),
                        lists=dict(feature=dict(column_type='list', uuid='lists'), type='feature'),
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            )
        ]
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)

    def test_seq_edge(self):
        df = pd.DataFrame(
            [
                ['MI', '32453', '12-26-2022'],
                ['CA', '', '12-28-2022'],
                ['', None, '12-28-2022'],
                ['MA', '12214', '12-31-2022'],
                ['PA', '', '1-2-2023'],
                ['TX', '75001', '1-2-2023'],
                ['CT', '06902', '12-24-2022'],
                ['NY', '10001', '12-25-2022'],
                [None, '', '12-30-2022'],
                ['CA', None, '12-30-2022'],
            ],
            columns=['state', 'location', 'timestamp'],
        )
        df['lists'] = pd.Series(
            [
                (np.nan, 2.0, 'string', '3'),
                ['not string?', True, True, 8, False, np.nan, None],
                None,
                ('not string?', True, True, 8, False, np.nan, None),
                [],
                '[\'not string?\'   ,  True, True , 8   , False  , np.nan, None]',
                None,
                '(\'not string?\'   ,  True, True , 8   , False  , np.nan, None)',
                tuple(),
                None,
            ]
        )
        column_types = {
            'state': 'category',
            'location': 'zip_code',
            'timestamp': 'datetime',
            'lists': 'list',
        }
        statistics = {
            'state/count': 7,
            'state/count_distinct': 6,
            'state/null_value_rate': 0.4,
            'state/max_null_seq': 1,
            'state/mode_ratio': 2 / 7,
            'location/count': 4,
            'location/count_distinct': 4,
            'location/null_value_rate': 0.6,
            'location/max_null_seq': 3,
            'location/mode_ratio': 1 / 4,
            'timestamp/null_value_rate': 0,
            'timestamp/max_null_seq': 0,
            'timestamp/mode_ratio': 1 / 10,
            'lists/null_value_rate': 0.3,
            'lists/max_null_seq': 1,
            'lists/mode_ratio': 2 / 10,
            'is_timeseries': True,
            'timeseries_index': ['timestamp'],
        }
        expected_suggestions = [
            dict(
                title='Fill in missing values',
                message='Fill missing entries using the previously occurring '
                'entry in the timeseries.',
                action_payload=dict(
                    action_type='impute',
                    action_arguments=['state', 'location', 'lists'],
                    action_options=dict(strategy='sequential', timeseries_index=['timestamp']),
                    action_variables=dict(
                        state=dict(
                            feature=dict(column_type='category', uuid='state'), type='feature'
                        ),
                        location=dict(
                            feature=dict(column_type='zip_code', uuid='location'), type='feature'
                        ),
                        lists=dict(feature=dict(column_type='list', uuid='lists'), type='feature'),
                    ),
                    action_code='',
                    axis='column',
                    outputs=[],
                ),
                status='not_applied',
            )
        ]
        df = clean_dataframe(df, column_types, dropna=False)
        suggestions = ImputeValues(
            df,
            column_types,
            statistics,
        ).evaluate()
        self.assertEqual(expected_suggestions, suggestions)



from datetime import datetime as dt
from mage_ai.data_cleaner.cleaning_rules.fix_syntax_errors import FixSyntaxErrors
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.statistics.calculator import StatisticsCalculator
from mage_ai.tests.base_test import TestCase
import pandas as pd
import numpy as np


class FixSyntaxErrorsTest(TestCase):
    def test_evaluate(self):
        df = pd.DataFrame(
            [
                [
                    '-12.3%',
                    1000,
                    '11111',
                    'email@maile.com',
                    '12223334444',
                    '12/13/2014',
                    dt(2022, 6, 27),
                    'a@b.c',
                ],
                [
                    '$2.234',
                    1050,
                    '2222-2222',
                    'er34ee@int.co',
                    ' 1(000)-111-2222',
                    'not a time',
                    dt(2022, 6, 27),
                    'b@c.d',
                ],
                [
                    'eieio',
                    np.nan,
                    '09876',
                    None,
                    '12345678901234',
                    '4/27/2019 12:34:45',
                    dt(2022, 6, 27),
                    'c@d.e',
                ],
                ['-4', 1150, None, 'email@email@email.com', 'qqqqqqqqqqq', None, dt(2022, 6, 27)],
                [
                    'not a number',
                    1150,
                    '23423932423',
                    'eeeeeeeee',
                    '1234',
                    '12-24-2022',
                    dt(2022, 6, 27),
                    'd@e.f',
                ],
                [
                    None,
                    1150,
                    '234.3324',
                    'agoodemail@network.net',
                    '43213240089',
                    'is not time',
                    dt(2022, 6, 27),
                    'e@f.g',
                ],
            ],
            columns=[
                'number',
                'number_but_correct_type',
                'zipcode',
                'email',
                'phone_number',
                'date',
                'date_but_correct_type',
                'email_but_all_correct',
            ],
        )
        column_types = {
            'number': ColumnType.NUMBER,
            'number_but_correct_type': ColumnType.NUMBER,
            'zipcode': ColumnType.ZIP_CODE,
            'email': ColumnType.EMAIL,
            'phone_number': ColumnType.PHONE_NUMBER,
            'date': ColumnType.DATETIME,
            'date_but_correct_type': ColumnType.DATETIME,
            'email_but_all_correct': ColumnType.DATETIME,
        }
        statistics = {
            'number/invalid_value_rate': 3 / 7,
            'number_but_correct_type/invalid_value_rate': 0 / 7,
            'zipcode/invalid_value_rate': 3 / 7,
            'email/invalid_value_rate': 3 / 7,
            'phone_number/invalid_value_rate': 4 / 7,
            'date/invalid_value_rate': 4 / 7,
            'date_but_correct_type/invalid_value_rate': 0 / 7,
            'email_but_all_correct/invalid_value_rate': 0 / 7,
        }
        result = FixSyntaxErrors(
            df,
            column_types,
            statistics,
        ).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Fix syntax errors',
                    message='Fix syntactical errors to reduce the amount of noise in the data.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='fix_syntax_errors',
                        action_arguments=[
                            'number',
                            'zipcode',
                            'email',
                            'phone_number',
                            'date',
                        ],
                        action_code='',
                        action_options={},
                        action_variables=dict(
                            number=dict(
                                feature=dict(column_type='number', uuid='number'), type='feature'
                            ),
                            zipcode=dict(
                                feature=dict(column_type='zip_code', uuid='zipcode'), type='feature'
                            ),
                            email=dict(
                                feature=dict(column_type='email', uuid='email'), type='feature'
                            ),
                            phone_number=dict(
                                feature=dict(column_type='phone_number', uuid='phone_number'),
                                type='feature',
                            ),
                            date=dict(
                                feature=dict(column_type='datetime', uuid='date'), type='feature'
                            ),
                        ),
                        axis='column',
                        outputs=[],
                    ),
                )
            ],
        )



from mage_ai.data_cleaner.column_types.column_type_detector import find_syntax_errors
from mage_ai.data_cleaner.column_types.constants import NUMBER_TYPES, ColumnType
from mage_ai.data_cleaner.estimators.outlier_removal import OutlierRemover
from mage_ai.data_cleaner.transformer_actions.action_code import query_with_action_code
from mage_ai.data_cleaner.transformer_actions.constants import (
    CONSTANT_IMPUTATION_DEFAULTS,
    CURRENCY_SYMBOLS,
    INVALID_VALUE_PLACEHOLDERS,
    ImputationStrategy,
)
from mage_ai.data_cleaner.transformer_actions.custom_action import execute_custom_action
from mage_ai.data_cleaner.transformer_actions.helpers import (
    convert_col_type,
    get_column_type,
    get_time_window_str,
)
from mage_ai.data_cleaner.transformer_actions.udf.base import execute_udf
from mage_ai.data_cleaner.transformer_actions.utils import clean_column_name, generate_string_cols
from keyword import iskeyword
import logging
import pandas as pd
import numpy as np


logger = logging.getLogger(__name__)


def add_column(df, action, **kwargs):
    col = action['outputs'][0]['uuid']
    col_type = action['outputs'][0]['column_type']
    udf = action['action_options'].get('udf')
    if udf is None:
        return df
    if len(action['action_arguments']) == 0:
        return df
    df_copy = df.copy()
    df_copy[col] = execute_udf(
        udf,
        df,
        action.get('action_arguments'),
        action.get('action_code'),
        action.get('action_options'),
        kwargs,
    )
    df_copy[col] = convert_col_type(df_copy[col], col_type)
    return df_copy


def average(df, action, **kwargs):
    return __agg(df, action, 'mean')


def count(df, action, **kwargs):
    return __groupby_agg(df, action, 'count')


def count_distinct(df, action, **kwargs):
    return __groupby_agg(df, action, 'nunique')


def clean_column_names(df, action, **kwargs):
    columns = action['action_arguments']
    mapping = {col: clean_column_name(col) for col in columns}
    return df.rename(columns=mapping)


def custom(df, action, **kwargs):
    return execute_custom_action(df, action, **kwargs)


def diff(df, action, **kwargs):
    if len(action['action_arguments']) == 0:
        return df
    output_col = action['outputs'][0]['uuid']
    df[output_col] = df[action['action_arguments'][0]].diff()
    return df


def first(df, action, **kwargs):
    return __agg(df, action, 'first')


def fix_syntax_errors(df, action, **kwargs):
    action_variables = action['action_variables']
    columns = action['action_arguments']
    for column in generate_string_cols(df, columns):
        dtype = action_variables[column]['feature']['column_type']
        mask = find_syntax_errors(df[column], dtype)
        df.loc[mask, column] = INVALID_VALUE_PLACEHOLDERS[dtype]
    return df


def impute(df, action, **kwargs):
    action_variables = action['action_variables']
    columns = action['action_arguments']
    action_options = action['action_options']
    strategy = action_options.get('strategy')
    value = action_options.get('value')

    empty_string_pattern = r'^\s*$'
    df[columns] = df[columns].replace(empty_string_pattern, np.nan, regex=True)
    ctypes = [action_variables[column]['feature']['column_type'] for column in columns]

    if strategy == ImputationStrategy.AVERAGE:
        df[columns] = df[columns].fillna(df[columns].astype(float).mean(axis=0))
    elif strategy == ImputationStrategy.CONSTANT:
        if value is None:
            value = pd.Series([CONSTANT_IMPUTATION_DEFAULTS[dtype] for dtype in ctypes])
            value.index = pd.Index(columns)
        df[columns] = df[columns].fillna(value)
    elif strategy == ImputationStrategy.MEDIAN:
        df[columns] = df[columns].fillna(df[columns].astype(float).median(axis=0))
    elif strategy == ImputationStrategy.MODE:
        if ColumnType.LIST in ctypes:
            for column, dtype in zip(columns, ctypes):
                mode = df[column].mode().iloc[0]
                if dtype == ColumnType.LIST:
                    df[column] = df[column].apply(
                        lambda element: element if element not in [None, np.nan] else mode
                    )
                else:
                    df[columns] = df[columns].fillna(mode)
        else:
            df[columns] = df[columns].fillna(df[columns].mode(axis=0).iloc[0])
    elif strategy == ImputationStrategy.COLUMN:
        replacement_df = pd.DataFrame({col: df[value] for col in columns})
        df[columns] = df[columns].fillna(replacement_df)
    elif strategy == ImputationStrategy.SEQ:
        timeseries_cols = action_options.get('timeseries_index')
        df = df.sort_values(by=timeseries_cols, axis=0)
        df[columns] = df[columns].fillna(method='ffill')
    elif strategy == ImputationStrategy.RANDOM:
        for column, dtype in zip(columns, ctypes):
            invalid_idx = df[df[column].isna()].index
            valid_idx = df[df[column].notna()].index
            if len(invalid_idx) == len(df[column]):
                raise Exception(
                    f'Random impute has no values to sample from in column \'{column}\''
                )
            sample = df.loc[valid_idx, column].sample(len(invalid_idx), replace=True)
            sample.index = invalid_idx
            if dtype == ColumnType.LIST:
                for idx, sample_value in zip(invalid_idx, sample):
                    df.at[idx, column] = sample_value
            else:
                df.loc[invalid_idx, column] = sample
    elif value is not None:
        df[columns] = df[columns].fillna(value)
    else:
        raise Exception('Require a valid strategy or value')

    for col in columns:
        col_type = get_column_type(col, action)
        df[col] = convert_col_type(df[col], col_type)
    return df


def max(df, action, **kwargs):
    return __agg(df, action, 'max')


def median(df, action, **kwargs):
    return __agg(df, action, 'median')


def min(df, action, **kwargs):
    return __agg(df, action, 'min')


def reformat(df, action, **kwargs):
    columns = action['action_arguments']
    options = action['action_options']
    reformat_action = options['reformat']
    df.loc[:, columns] = df[columns].replace('^\s*$', np.nan, regex=True)

    if reformat_action == 'caps_standardization':
        capitalization = options['capitalization']
        for column in generate_string_cols(df, columns):
            if capitalization == 'uppercase':
                df.loc[:, column] = df[columns][column].str.upper()
            else:
                df.loc[:, column] = df[columns][column].str.lower()
    elif reformat_action == 'currency_to_num':
        for column in generate_string_cols(df, columns):
            clean_col = df[column].replace(CURRENCY_SYMBOLS, '', regex=True)
            clean_col = clean_col.replace('\s', '', regex=True)
            clean_col = clean_col.replace('^\s*$', np.nan, regex=True)
            try:
                df.loc[:, column] = clean_col.astype(float)
            except ValueError:
                logger.warn(
                    f'Currency conversion applied on non-numerical column \'{column}\''
                    ': no action taken'
                )
    elif reformat_action == 'date_format_conversion':
        for column in columns:
            clean_col = df[column]
            dropped = clean_col.dropna(axis=0)
            exact_dtype = type(dropped.iloc[0]) if len(dropped) > 0 else None
            if exact_dtype is str:
                clean_col = clean_col.str.replace(r'[\,\s\t]+', ' ')
                clean_col = clean_col.str.replace(
                    r'\s*([\/\\\-\.]+)\s*', lambda group: group.group(1)[0]
                )
                clean_col = clean_col.str.lower()
            df.loc[:, column] = pd.to_datetime(
                clean_col, infer_datetime_format=True, errors='coerce'
            )

    return df


def remove_column(df, action, **kwargs):
    cols = action['action_arguments']
    original_columns = df.columns
    drop_columns = [col for col in cols if col in original_columns]

    return df.drop(columns=drop_columns)


def remove_outliers(df, action, **kwargs):
    cols = set(action['action_arguments'])
    numeric_df = df[cols].copy()
    for column in numeric_df.columns:
        dtype = action['action_variables'][column]['feature']['column_type']
        if dtype in NUMBER_TYPES:
            numeric_df.loc[:, column] = numeric_df.loc[:, column].astype(float)
        else:
            numeric_df.drop(column, axis=1, inplace=True)
    outlier_mask = numeric_df.notna().all(axis=1)
    numeric_df = numeric_df.dropna(axis=0)
    if numeric_df.size == 0:
        return df

    method = action['action_options']['method']
    remover = OutlierRemover(method=method)
    notna_outlier_mask = remover.fit_transform(numeric_df.to_numpy())
    # This code maps the outlier mask on a subset of data back to the mask on the entire data
    notna_outlier_mask = pd.Series(notna_outlier_mask, dtype='bool')
    notna_outlier_mask.index = outlier_mask[outlier_mask].index
    outlier_mask[outlier_mask] = notna_outlier_mask
    outlier_mask = outlier_mask.astype(bool)
    return df[~outlier_mask]


def last(df, action, **kwargs):
    return __agg(df, action, 'last')


def select(df, action, **kwargs):
    return df[action['action_arguments']]


def shift_down(df, action, **kwargs):
    if len(action['action_arguments']) == 0:
        return df

    output_col = action['outputs'][0]['uuid']
    action_options = action.get('action_options', {})
    groupby_columns = action_options.get('groupby_columns')
    periods = action_options.get('periods', 1)
    if groupby_columns is not None:
        df[output_col] = df.groupby(groupby_columns)[action['action_arguments'][0]].shift(periods)
    else:
        df[output_col] = df[action['action_arguments'][0]].shift(periods)
    return df


def shift_up(df, action, **kwargs):
    if len(action['action_arguments']) == 0:
        return df
    output_col = action['outputs'][0]['uuid']
    df[output_col] = df[action['action_arguments'][0]].shift(-1)
    return df


def sum(df, action, **kwargs):
    return __agg(df, action, 'sum')


def __agg(df, action, agg_method):
    if len(action['action_arguments']) == 0:
        return df

    if action['action_options'].get('groupby_columns'):
        return __groupby_agg(df, action, agg_method)
    else:
        output_col = action['outputs'][0]['uuid']
        df[output_col] = df[action['action_arguments'][0]].agg(agg_method)
        return df


def __column_mapping(action):
    return dict(zip(action['action_arguments'], [o['uuid'] for o in action['outputs']]))


# Filter by timestamp_feature_a - window <= timestamp_feature_b <= timestamp_feature_a
def __filter_df_with_time_window(df, action):
    if len(action['action_arguments']) == 0:
        return df

    action_options = action['action_options']
    time_window_keys = ['timestamp_feature_a', 'timestamp_feature_b', 'window']
    if all(k in action_options for k in time_window_keys):
        window_in_seconds = action_options['window']
        df_time_diff = (
            pd.to_datetime(df[action_options['timestamp_feature_a']], utc=True)
            - pd.to_datetime(df[action_options['timestamp_feature_b']], utc=True)
        ).dt.total_seconds()
        if window_in_seconds > 0:
            df_time_diff_filtered = df_time_diff[
                (df_time_diff <= window_in_seconds) & (df_time_diff >= 0)
            ]
        else:
            df_time_diff_filtered = df_time_diff[
                (df_time_diff >= window_in_seconds) & (df_time_diff <= 0)
            ]
        df_filtered = df.loc[df_time_diff_filtered.index]
        time_window = get_time_window_str(window_in_seconds)
    else:
        df_filtered = df
        time_window = None
    return df_filtered, time_window


def __groupby_agg(df, action, agg_method):
    if len(action['action_arguments']) == 0:
        return df

    df_filtered, _ = __filter_df_with_time_window(df, action)
    action_code = action.get('action_code')
    if action_code is not None and action_code != '':
        df_filtered = query_with_action_code(df_filtered, action_code, {'original_df': df_filtered})
    action_options = action['action_options']
    df_agg = df_filtered.groupby(action_options['groupby_columns'],)[
        action['action_arguments']
    ].agg(agg_method)
    return df.merge(
        df_agg.rename(columns=__column_mapping(action)),
        on=action_options['groupby_columns'],
        how='left',
    )



from mage_ai.cli.utils import parse_runtime_variables
import argparse
import os


def main():
    import sys

    try:
        command = sys.argv[1]
    except IndexError:
        command = 'help'

    if command == 'help':
        print("""Usage:
    mage <command> 

Commands:
    init <project_path>                 Initialize Mage project.
      args:
        project_path                    path of the Mage project to be created.

    start <project_path>                Start Mage server and UI.
      args:
        project_path                    path of the Mage project to be loaded.
      options:
        --host <host>                   specify the host, defaults to localhost
        --port <port>                   specify the port, defaults to 6789

    run <project_path> <pipeline_uuid>  Run pipeline.
      args:
        project_path                    path of the Mage project that contains the pipeline.
        pipeline_uuid                   uuid of the pipeline to be run.
      options:
        --runtime-vars [key value]...   specify runtime variables. These will overwrite the pipeline global variables.   

    test <project_path> <pipeline_uuid> Run pipeline and output tests.
      args:
        project_path                    path of the Mage project that contains the pipeline.
        pipeline_uuid                   uuid of the pipeline to be run and tested.
      options:
        --runtime-vars [key value]...   specify runtime variables. These will overwrite the pipeline global variables.

    create_spark_cluster <project_path>
      args:
        project_path                    path of the Mage project that contains the EMR config.
        """)
    elif command == 'init':
        from mage_ai.data_preparation.repo_manager import init_repo

        repo_path = os.path.join(os.getcwd(), sys.argv[2])
        init_repo(repo_path)
    elif command == 'start':
        from mage_ai.server.server import start_server

        parser = argparse.ArgumentParser()
        parser.add_argument('repo_path', metavar='project_path', type=str)
        parser.add_argument('--host', nargs='?', type=str)
        parser.add_argument('--port', nargs='?', type=int)

        args = dict()
        if len(sys.argv) >= 3:
            args = vars(parser.parse_args(sys.argv[2:]))
            repo_path = args['repo_path']
        else:
            repo_path = os.getcwd()

        start_server(
            host=args.get('host'),
            port=args.get('port'),
            project=repo_path,
        )
    elif command == 'run' or command == 'test':
        from mage_ai.data_preparation.executors.executor_factory import ExecutorFactory
        from mage_ai.data_preparation.models.pipeline import Pipeline
        from mage_ai.data_preparation.repo_manager import set_repo_path
        from mage_ai.data_preparation.variable_manager import get_global_variables
        from mage_ai.shared.hash import merge_dict

        parser = argparse.ArgumentParser(description='Run pipeline.')
        parser.add_argument('repo_path', metavar='project_path', type=str)
        parser.add_argument('pipeline_uuid', type=str)
        parser.add_argument('--block_uuid', nargs='?', type=str)
        parser.add_argument('--execution_partition', nargs='?', type=str)
        parser.add_argument('--executor_type', nargs='?', type=str)
        parser.add_argument('--callback_url', nargs='?', type=str)
        parser.add_argument('--block_run_id', nargs='?', type=int)
        parser.add_argument('--runtime-vars', nargs="+")
        parser.add_argument('--skip-sensors', type=bool)

        args = vars(parser.parse_args(sys.argv[2:]))
        project_path = args['repo_path']
        pipeline_uuid = args['pipeline_uuid']
        block_uuid = args.get('block_uuid')
        execution_partition = args.get('execution_partition')
        executor_type = args.get('executor_type')
        callback_url = args.get('callback_url')
        runtime_vars = args.get('runtime_vars')
        skip_sensors = args.get('skip_sensors', False)

        runtime_variables = dict()
        if runtime_vars is not None:
            runtime_variables = parse_runtime_variables(runtime_vars)

        project_path = os.path.abspath(project_path)
        set_repo_path(project_path)
        sys.path.append(os.path.dirname(project_path))
        pipeline = Pipeline(pipeline_uuid, repo_path=project_path)

        default_variables = get_global_variables(pipeline_uuid, repo_path=project_path)
        global_vars = merge_dict(default_variables, runtime_variables)

        if block_uuid is None:
            ExecutorFactory.get_pipeline_executor(pipeline).execute(
                analyze_outputs=False,
                global_vars=global_vars,
                run_sensors=not skip_sensors,
                run_tests=command == 'test',
                update_status=False,
            )
        else:
            ExecutorFactory.get_block_executor(
                pipeline,
                block_uuid,
                execution_partition=execution_partition,
                executor_type=executor_type,
            ).execute(
                analyze_outputs=False,
                callback_url=callback_url,
                global_vars=global_vars,
                update_status=False,
            )

    elif command == 'create_spark_cluster':
        from mage_ai.services.aws.emr.launcher import create_cluster

        project_path = os.path.abspath(sys.argv[2])
        create_cluster(project_path)
    else:
        print(f'Unknown command "{command}". Type "mage help" to see what commands are available.')


if __name__ == "__main__":
    main()



from datetime import datetime, timedelta
from mage_ai.shared.hash import group_by, merge_dict
from mage_ai.orchestration.db.models import BlockRun, PipelineRun
from sqlalchemy.orm import joinedload
from typing import Callable, Dict, List
import dateutil.parser
import enum


class MonitorStatsType(str, enum.Enum):
    PIPELINE_RUN_COUNT = 'pipeline_run_count'
    PIPELINE_RUN_TIME = 'pipeline_run_time'
    BLOCK_RUN_COUNT = 'block_run_count'
    BLOCK_RUN_TIME = 'block_run_time'


class MonitorStats:
    def get_stats(
        self,
        stats_type: MonitorStatsType,
        pipeline_uuid: str = None,
        start_time: str = None,
        end_time: str = None,
        **kwargs,
    ) -> Dict:
        if end_time is None:
            end_time = datetime.now()
        else:
            end_time = dateutil.parser.parse(end_time)
        if start_time is None:
            start_time = end_time - timedelta(days=30)
        else:
            start_time = dateutil.parser.parse(start_time)
        new_kwargs = merge_dict(dict(
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
        ), kwargs)
        if stats_type == MonitorStatsType.PIPELINE_RUN_COUNT:
            return self.get_pipeline_run_count(**new_kwargs)
        elif stats_type == MonitorStatsType.PIPELINE_RUN_TIME:
            return self.get_pipeline_run_time(**new_kwargs)
        elif stats_type == MonitorStatsType.BLOCK_RUN_COUNT:
            return self.get_block_run_count(**new_kwargs)
        elif stats_type == MonitorStatsType.BLOCK_RUN_TIME:
            return self.get_block_run_time(**new_kwargs)

    def get_pipeline_run_count(
        self,
        pipeline_uuid: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> Dict:
        pipeline_runs = self.__filter_pipeline_runs(
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )
        pipeline_runs = pipeline_runs.all()
        stats_by_schedule_id = dict()
        for p in pipeline_runs:
            if p.pipeline_schedule_id not in stats_by_schedule_id:
                stats_by_schedule_id[p.pipeline_schedule_id] = dict(
                    name=p.pipeline_schedule_name,
                    data=dict(),
                )
            created_at_formatted = p.created_at.strftime('%Y-%m-%d')
            data = stats_by_schedule_id[p.pipeline_schedule_id]['data']
            if created_at_formatted not in data:
                data[created_at_formatted] = dict()
            if p.status not in data[created_at_formatted]:
                data[created_at_formatted][p.status] = 1
            else:
                data[created_at_formatted][p.status] += 1
        return stats_by_schedule_id

    def get_pipeline_run_time(
        self,
        pipeline_uuid: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> Dict:
        pipeline_runs = self.__filter_pipeline_runs(
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )
        pipeline_runs = pipeline_runs.filter(PipelineRun.completed_at != None).all()
        pipeline_run_by_date = group_by(lambda p: p.created_at.strftime('%Y-%m-%d'), pipeline_runs)

        def __mean_runtime(pipeline_runs):
            runtime_list = [(p.completed_at - p.created_at).total_seconds()
                            for p in pipeline_runs if p.completed_at > p.created_at]
            if len(runtime_list) == 0:
                return 0
            return sum(runtime_list) / len(runtime_list)
        pipeline_run_time_by_date = {k: __mean_runtime(v) for k, v in pipeline_run_by_date.items()}
        return {pipeline_uuid: dict(name=pipeline_uuid, data=pipeline_run_time_by_date)}

    def get_block_run_count(
        self,
        pipeline_uuid: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> Dict:
        block_runs = self.__filter_block_runs(
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )
        block_runs = block_runs.all()

        def __stats_func(block_runs):
            count_by_status = dict()
            for b in block_runs:
                if b.status in count_by_status:
                    count_by_status[b.status] += 1
                else:
                    count_by_status[b.status] = 1
            return count_by_status

        return self.__cal_block_run_stats(block_runs, __stats_func)

    def get_block_run_time(
        self,
        pipeline_uuid: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> Dict:
        block_runs = self.__filter_block_runs(
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
            **kwargs,
        )
        block_runs = block_runs.filter(BlockRun.completed_at != None).all()

        def __stats_func(block_runs):
            runtime_list = [(b.completed_at - b.created_at).total_seconds()
                            for b in block_runs if b.completed_at > b.created_at]
            if len(runtime_list) == 0:
                return 0
            return sum(runtime_list) / len(runtime_list)

        return self.__cal_block_run_stats(block_runs, __stats_func)

    def __filter_pipeline_runs(
        self,
        pipeline_uuid: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs
    ) -> Dict:
        pipeline_runs = PipelineRun.query.options(joinedload(PipelineRun.pipeline_schedule))
        if pipeline_uuid is not None:
            pipeline_runs = pipeline_runs.filter(PipelineRun.pipeline_uuid == pipeline_uuid)
        if start_time is not None:
            pipeline_runs = pipeline_runs.filter(PipelineRun.created_at >= start_time)
        if end_time is not None:
            pipeline_runs = pipeline_runs.filter(PipelineRun.created_at <= end_time)
        pipeline_schedule_id = kwargs.get('pipeline_schedule_id')
        if pipeline_schedule_id is not None:
            pipeline_runs = pipeline_runs.filter(
                PipelineRun.pipeline_schedule_id == int(pipeline_schedule_id)
            )

        return pipeline_runs

    def __filter_block_runs(
        self,
        pipeline_uuid: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> Dict:
        block_runs = BlockRun.query
        if pipeline_uuid is not None:
            block_runs = (
                block_runs.
                join(
                    PipelineRun,
                    PipelineRun.id == BlockRun.pipeline_run_id,
                ).
                filter(
                    PipelineRun.pipeline_uuid == pipeline_uuid,
                )
            )
        if start_time is not None:
            block_runs = block_runs.filter(BlockRun.created_at >= start_time)
        if end_time is not None:
            block_runs = block_runs.filter(BlockRun.created_at <= end_time)

        pipeline_schedule_id = kwargs.get('pipeline_schedule_id')
        if pipeline_schedule_id is not None:
            block_runs = block_runs.filter(
                PipelineRun.pipeline_schedule_id == int(pipeline_schedule_id)
            )
        return block_runs

    def __cal_block_run_stats(
        self,
        block_runs: List[BlockRun],
        stats_func: Callable
    ) -> Dict:
        block_runs_by_uuid = group_by(lambda b: b.block_uuid, block_runs)
        block_run_stats = dict()
        for uuid, sub_block_runs in block_runs_by_uuid.items():
            sub_block_runs_by_date = group_by(
                lambda b: b.created_at.strftime('%Y-%m-%d'),
                sub_block_runs,
            )
            sub_block_runs_stats = {k: stats_func(v) for k, v in sub_block_runs_by_date.items()}
            block_run_stats[uuid] = dict(
                name=uuid,
                data=sub_block_runs_stats,
            )
        return block_run_stats



from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF
import numpy as np
import pandas as pd


class Difference(BaseUDF):
    def execute(self):
        col1 = self.arguments[0]
        column_type = self.options.get('column_type', self.kwargs.get('column_types', {}).get(col1))
        if len(self.arguments) > 1:
            col2 = self.arguments[1]
            return self.__difference_between_columns(
                self.df[col1],
                self.df[col2],
                column_type=column_type,
                options=self.options,
            )
        elif self.options.get('value') is not None:
            return self.__subtract_value(
                self.df[col1],
                self.options['value'],
                column_type=column_type,
                options=self.options,
            )
        raise Exception('Require second column or a value to minus.')

    def __difference_between_columns(self, column1, column2, column_type=None, options={}):
        if column_type == ColumnType.DATETIME:
            time_unit = options.get('time_unit', 'd')
            return (pd.to_datetime(column1, utc=True) - pd.to_datetime(column2, utc=True)).dt.days
        return column1 - column2

    def __subtract_value(self, original_column, value, column_type=None, options={}):
        if column_type == ColumnType.DATETIME:
            time_unit = options.get('time_unit', 'd')
            return (
                pd.to_datetime(original_column, utc=True) - pd.to_timedelta(value, unit=time_unit)
            ).dt.strftime('%Y-%m-%d %H:%M:%S')
        return original_column - value



from mage_ai.data_cleaner.analysis.constants import (
    CHART_TYPE_BAR_HORIZONTAL,
    CHART_TYPE_LINE_CHART,
    CHART_TYPE_HISTOGRAM,
    DATA_KEY_SCATTER_PLOT,
    DATA_KEY_SCATTER_PLOT_LABELS,
    DATA_KEY_TIME_SERIES,
    LABEL_TYPE_RANGE,
)
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.estimators.encoders import MultipleColumnLabelEncoder
from mage_ai.data_cleaner.shared.utils import clean_series
import math
import numpy as np
import pandas as pd

DD_KEY = 'lambda.analysis_charts'
BUCKETS = 40
SCATTER_PLOT_SAMPLE_COUNT = 200
SCATTER_PLOT_CATEGORY_LIMIT = 10
TIME_SERIES_BUCKETS = 40


def increment(metric, tags={}):
    pass


def build_buckets(min_value, max_value, max_buckets, column_type):
    diff = max_value - min_value
    total_interval = 1 + diff
    bucket_interval = total_interval / max_buckets
    number_of_buckets = max_buckets

    if diff < 0.0001:
        diff = 0

    is_integer = False
    parts = str(diff).split('.')
    if len(parts) == 1:
        is_integer = True
    else:
        is_integer = int(parts[1]) == 0

    if ColumnType.NUMBER == column_type and total_interval <= max_buckets and is_integer:
        number_of_buckets = int(total_interval)
        bucket_interval = 1
    elif bucket_interval > 1:
        bucket_interval = math.ceil(bucket_interval)
    else:
        bucket_interval = round(bucket_interval * 100, 1) / 100

    buckets = []
    for i in range(number_of_buckets):
        min_v = min_value + (i * bucket_interval)
        max_v = min_value + ((i + 1) * bucket_interval)
        buckets.append(
            dict(
                max_value=max_v,
                min_value=min_v,
                values=[],
            )
        )

    return buckets, bucket_interval


def build_histogram_data(col1, series, column_type):
    increment(f'{DD_KEY}.build_histogram_data.start', dict(feature_uuid=col1))

    max_value = series.max()
    min_value = series.min()

    buckets, bucket_interval = build_buckets(min_value, max_value, BUCKETS, column_type)

    if bucket_interval == 0:
        return

    bins = [b['min_value'] for b in buckets] + [buckets[-1]['max_value']]
    count, _ = np.histogram(series, bins=bins)

    x = []
    y = []

    for idx, bucket in enumerate(buckets):
        x.append(
            dict(
                max=bucket['max_value'],
                min=bucket['min_value'],
            )
        )
        y.append(dict(value=count[idx]))

    increment(f'{DD_KEY}.build_histogram_data.succeeded', dict(feature_uuid=col1))

    return dict(
        type=CHART_TYPE_HISTOGRAM,
        x=x,
        x_metadata=dict(
            label_type=LABEL_TYPE_RANGE,
        ),
        y=y,
    )


def build_correlation_data(df):
    charts = dict()
    df_corr = df.corr()
    columns = df_corr.columns
    for col1 in columns:
        x = []
        y = []
        for col2 in columns:
            if col1 != col2:
                value = df_corr[col1].get(col2, None)
                if value is not None:
                    x.append(dict(label=col2))
                    y.append(dict(value=value))
        charts[col1] = [dict(
            type=CHART_TYPE_BAR_HORIZONTAL,
            x=x,
            y=y,
        )]
    return charts


def build_time_series_data(df, features, datetime_column):
    # print(feature, datetime_column)

    datetimes = df[datetime_column].dropna()
    if datetimes.size <= 1:
        return

    # datetimes = pd.to_datetime(datetimes, infer_datetime_format=True, errors='coerce')

    min_value_datetime = datetimes.min().timestamp()
    max_value_datetime = datetimes.max().timestamp()

    buckets, bucket_interval = build_buckets(
        min_value_datetime,
        max_value_datetime,
        TIME_SERIES_BUCKETS,
        ColumnType.DATETIME,
    )

    x = []
    y_dict = dict()

    df_copy = df.copy()
    df_copy[datetime_column] = datetimes.view(int) / 10**9

    for bucket in buckets:
        max_value = bucket['max_value']
        min_value = bucket['min_value']

        df_filtered = df_copy[
            (df_copy[datetime_column] >= min_value) & (df_copy[datetime_column] < max_value)
        ]

        x.append(
            dict(
                max=max_value,
                min=min_value,
            )
        )

        series_count = df_filtered.shape[0]
        for f in features:
            col = f['uuid']
            column_type = f['column_type']
            if col not in y_dict:
                y_dict[col] = []

            series = df_filtered[col]
            # series_cleaned = clean_series(series, column_type, dropna=False)
            series_cleaned = series
            series_non_null = series_cleaned.dropna()
            non_null_count = series_non_null.size

            y_data = dict(
                count=non_null_count,
                count_distinct=series_non_null.nunique(),
                null_value_rate=0
                if series_count == 0
                else (series_count - non_null_count) / series_count,
            )

            if column_type in [ColumnType.NUMBER, ColumnType.NUMBER_WITH_DECIMALS]:
                if len(series_non_null) == 0:
                    average = 0
                else:
                    average = series_non_null.mean()
                y_data.update(
                    dict(
                        average=average,
                        max=series_non_null.max(),
                        median=series_non_null.median(),
                        min=series_non_null.min(),
                        sum=series_non_null.sum(),
                    )
                )
            elif column_type in [
                ColumnType.CATEGORY,
                ColumnType.CATEGORY_HIGH_CARDINALITY,
                ColumnType.TRUE_OR_FALSE,
            ]:
                value_counts = series_non_null.value_counts()
                if len(value_counts.index):
                    value_counts_top = value_counts.iloc[:12]
                    mode = value_counts_top.index[0]
                    y_data.update(
                        dict(
                            mode=mode,
                            value_counts=value_counts_top.to_dict(),
                        )
                    )

            y_dict[col].append(y_data)
    charts = dict()
    for f, y in y_dict.items():
        charts[f] = dict(
            type=CHART_TYPE_LINE_CHART,
            x=x,
            x_metadata=dict(
                label=datetime_column,
                label_type=LABEL_TYPE_RANGE,
            ),
            y=y,
        )
    return charts


def build_overview_data(
    df,
    datetime_features,
    numeric_features,
):
    increment(f'{DD_KEY}.build_overview_data.start')

    time_series = []
    df_copy = df.copy()

    for feature in datetime_features:
        column_type = feature['column_type']
        datetime_column = feature['uuid']
        tags = dict(datetime_column=datetime_column)
        increment(f'{DD_KEY}.build_overview_time_series.start', tags)

        if df_copy[datetime_column].count() <= 1:
            continue

        df_copy[datetime_column] = pd.to_datetime(
            df[datetime_column], infer_datetime_format=True, errors='coerce'
        )
        df_copy[datetime_column] = df_copy[datetime_column].view(int) / 10**9

        min_value1 = df_copy[datetime_column].min()
        max_value1 = df_copy[datetime_column].max()
        buckets, bucket_interval = build_buckets(
            min_value1, max_value1, TIME_SERIES_BUCKETS, column_type
        )

        x = []
        y = []

        for bucket in buckets:
            max_value = bucket['max_value']
            min_value = bucket['min_value']

            df_filtered = df_copy[
                (df_copy[datetime_column] >= min_value) & (df_copy[datetime_column] < max_value)
            ]

            x.append(
                dict(
                    max=max_value,
                    min=min_value,
                )
            )

            y.append(
                dict(
                    count=len(df_filtered.index),
                )
            )

        time_series.append(
            dict(
                type=CHART_TYPE_LINE_CHART,
                x=x,
                x_metadata=dict(
                    label=datetime_column,
                    label_type=LABEL_TYPE_RANGE,
                ),
                y=y,
            )
        )

        increment(f'{DD_KEY}.build_overview_time_series.succeeded', tags)

    """
    Build sample data for scatter plot. Sample data consits of two parts:
    1. Numeric features
    2. Low cardinality categorical features
    """
    if df.shape[0] > SCATTER_PLOT_SAMPLE_COUNT:
        df_sample = df.sample(SCATTER_PLOT_SAMPLE_COUNT).copy()
    else:
        df_sample = df.copy()

    df_sample_numeric = df_sample[numeric_features].dropna(axis=1, how='all')
    """
    Calculate low cardinality categorical features:
    1. unique count <= SCATTER_PLOT_CATEGORY_LIMIT and unique count > 1
    2. count > SCATTER_PLOT_SAMPLE_COUNT / 2
    """
    non_numeric_features = list(set(df.columns) - set(numeric_features))
    non_numeric_nuniques = df_sample[non_numeric_features].nunique()
    non_numeric_nuniques_filtered = non_numeric_nuniques[
        (non_numeric_nuniques <= SCATTER_PLOT_CATEGORY_LIMIT) & (non_numeric_nuniques > 1)
    ]
    non_numeric_counts = df_sample[non_numeric_features].count()
    sample_count = df_sample.shape[0]
    non_numeric_counts_filtered = non_numeric_counts[non_numeric_counts > sample_count / 2]
    eligible_category_features = set(non_numeric_nuniques_filtered.index) & set(
        non_numeric_counts_filtered.index
    )
    if len(eligible_category_features) > 0:
        encoder = MultipleColumnLabelEncoder(input_type=str)
        df_sample_category = encoder.fit_transform(df_sample[eligible_category_features])
        class_mappings = {k: list(e.label_classes()) for k, e in encoder.encoders.items()}
        df_sample_filtered = pd.concat([df_sample_numeric, df_sample_category], axis=1)
    else:
        class_mappings = dict()
        df_sample_filtered = df_sample_numeric

    increment(f'{DD_KEY}.build_overview_data.succeeded')

    return {
        DATA_KEY_TIME_SERIES: time_series,
        DATA_KEY_SCATTER_PLOT: df_sample_filtered.to_dict('list'),
        DATA_KEY_SCATTER_PLOT_LABELS: class_mappings,
    }



from mage_ai.orchestration.notification.config import NotificationConfig
from mage_ai.services.email.email import send_email
from mage_ai.services.slack.slack import send_slack_message
import os


class NotificationSender:
    def __init__(self, config: NotificationConfig):
        self.config = config

    def send(
        self,
        message: str = None,
        email_subject: str = None,
        email_content: str = None,
    ) -> None:
        if message is None:
            return
        if self.config.slack_config is not None and self.config.slack_config.is_valid:
            send_slack_message(self.config.slack_config, message)
        if self.config.email_config is not None and email_subject is not None:
            send_email(
                self.config.email_config,
                subject=email_subject,
                message=email_content or message,
            )

    def send_pipeline_run_success_message(self, pipeline, pipeline_run) -> None:
        message = (
            f'Successfully ran Pipeline `{pipeline.uuid}` '
            f'with Trigger {pipeline_run.pipeline_schedule.id} '
            f'`{pipeline_run.pipeline_schedule.name}` '
            f'at execution time `{pipeline_run.execution_date}`.'
        )
        email_content = f'{message}\n'
        if os.getenv('ENV') != 'production':
            email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                'to check pipeline run results and logs.'
        self.send(
            message=message,
            email_subject=f'Successfully ran Pipeline {pipeline.uuid}',
            email_content=email_content,
        )

    def send_pipeline_run_failure_message(self, pipeline, pipeline_run) -> None:
        message = (
            f'Failed to run Pipeline `{pipeline.uuid}` '
            f'with Trigger {pipeline_run.pipeline_schedule.id} '
            f'`{pipeline_run.pipeline_schedule.name}` '
            f'at execution time `{pipeline_run.execution_date}`.'
        )
        email_content = f'{message}\n'
        if os.getenv('ENV') != 'production':
            email_content += f'Open {self.__pipeline_run_url(pipeline, pipeline_run)} '\
                                'to check pipeline run results and logs.'
        self.send(
            message=message,
            email_subject=f'Failed to run Mage pipeline {pipeline.uuid}',
            email_content=email_content,
        )

    def __pipeline_run_url(self, pipeline, pipeline_run):
        return f'http://localhost:6789/pipelines/{pipeline.uuid}/triggers/'\
                f'{pipeline_run.pipeline_schedule_id}'



from .base import BaseHandler
from mage_ai.orchestration.monitor.monitor_stats import MonitorStats


class ApiMonitorStatsHandler(BaseHandler):
    def get(self, stats_type):
        pipeline_uuid = self.get_argument('pipeline_uuid', None)
        start_time = self.get_argument('start_time', None)
        end_time = self.get_argument('end_time', None)
        pipeline_schedule_id = self.get_argument('pipeline_schedule_id', None)
        stats = MonitorStats().get_stats(
            stats_type,
            pipeline_uuid=pipeline_uuid,
            start_time=start_time,
            end_time=end_time,
            pipeline_schedule_id=pipeline_schedule_id,
        )
        self.write(dict(
            monitor_stats=dict(
                stats_type=stats_type,
                stats=stats,
            ),
        ))



from jinja2 import Template
from mage_ai.data_preparation.models.constants import BlockLanguage
import re


def interpolate_input(block, query, replace_func=None):
    def __replace_func(db, schema, tn):
        if replace_func:
            return replace_func(db, schema, tn)

        return f'{schema}.{tn}'

    for idx, upstream_block in enumerate(block.upstream_blocks):
        matcher1 = '{} df_{} {}'.format('{{', idx + 1, '}}')

        if BlockLanguage.SQL == upstream_block.type:
            configuration = upstream_block.configuration
        else:
            configuration = block.configuration

        database = configuration.get('data_provider_database', '')
        schema = configuration.get('data_provider_schema', '')

        query = re.sub(
            '{}[ ]*df_{}[ ]*{}'.format('\{\{', idx + 1, '\}\}'),
            __replace_func(database, schema, upstream_block.table_name),
            query,
        )

        query = query.replace(
            f'{matcher1}',
            __replace_func(database, schema, upstream_block.table_name),
        )

    return query


def interpolate_vars(query, global_vars=dict()):
    return Template(query).render(**global_vars)



from functools import reduce
import math
import re


def dig(obj_arg, arr_or_string):
    if type(arr_or_string) is str:
        arr_or_string = arr_or_string.split('.')
    arr = list(map(str.strip, arr_or_string))
    def _build(obj, key):
        tup = re.split(r'\[(\d+)\]$', key)
        if len(tup) >= 2:
            key, index = filter(lambda x: x, tup)
            if key and index:
                return obj[key][int(index)]
            elif index:
                return obj[int(index)]
        elif obj:
            return obj.get(key)
        else:
            return obj
    return reduce(_build, arr, obj_arg)


def flatten(input_data):
    final_data = {}

    for k1, v1 in input_data.items():
        if type(v1) is dict:
            for k2, v2 in v1.items():
                if type(v2) is dict:
                    for k3, v3 in v2.items():
                        final_data[f'{k1}_{k2}_{k3}'] = v3
                else:
                    final_data[f'{k1}_{k2}'] = v2
        else:
            final_data[k1] = v1

    return final_data


def ignore_keys(d, keys):
    d_keys = d.keys()
    d2 = d.copy()
    for key in keys:
        if key in d_keys:
            d2.pop(key)
    return d2


def ignore_keys_with_blank_values(d):
    d2 = d.copy()
    for key, value in d.items():
        if not value:
            d2.pop(key)
    return d2


def extract(d, keys):
    def _build(obj, key):
        if key in d:
            val = d[key]
        if val is not None:
            obj[key] = val
        return obj
    return reduce(_build, keys, {})


def extract_arrays(input_data):
    arr = []
    for k, v in input_data.items():
        if type(v) is list:
            arr.append(v)
    return arr


def group_by(func, arr):
    def _build(obj, item):
        val = func(item)
        if not obj.get(val):
            obj[val] = []
        obj[val].append(item)
        return obj
    return reduce(_build, arr, {})


def index_by(func, arr):
    obj = {}
    for item in arr:
        key = func(item)
        obj[key] = item
    return obj


def merge_dict(a, b):
    c = a.copy()
    c.update(b)
    return c


def replace_dict_nan_value(d):
    def _replace_nan_value(v):
        if type(v) == float and math.isnan(v):
            return None
        return v
    return {k: _replace_nan_value(v) for k, v in d.items()}



from io import BytesIO
from google.cloud import storage
from google.oauth2 import service_account
from mage_ai.io.base import BaseFile, FileFormat, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from pandas import DataFrame


class GoogleCloudStorage(BaseFile):
    """
    Handles data transfer between a Google Cloud Storage bucket and the Mage app. Supports loading files
    of any of the following types:
    - ".csv"
    - ".json"
    - ".parquet"

    If GOOGLE_APPLICATION_CREDENTIALS environment is set, no further arguments are needed other than those
    specified below. Otherwise, use the factory method `with_config` to construct
    the data loader using manually specified credentials.
    """

    def __init__(
        self,
        verbose=False,
        **kwargs,
    ) -> None:
        """
        Initializes data loader from a Google Cloud Storage bucket.

        To authenticate (and authorize) access to Google Cloud Storage, credentials must be provided.
        Below are the different ways to access those credentials:
        - Define the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to point to a
          service account key. In this case no other no other parameters need to be
          specified.
        - Manually pass in the `google.oauth2.service_account.Credentials` object with the
        keyword argument `credentials`
        - Manually pass in the path to the credentials with the keyword argument
        `path_to_credentials`.
        - Manually define the service key key-value set to use (such as a dictionary containing
        the same parameters as a service key) with the keyword argument `credentials_mapping`

        All keyword arguments except for `path_to_credentials` and `credentials_mapping` will be passed
        to the Google Cloud Storage client, accepting all other configuration settings there.
        """

        super().__init__(verbose=kwargs.get('verbose', True))

        if kwargs.get('verbose') is not None:
            kwargs.pop('verbose')

        credentials = kwargs.get('credentials')
        if credentials is None:
            if 'credentials_mapping' in kwargs:
                mapping_obj = kwargs.pop('credentials_mapping')
                if mapping_obj is not None:
                    credentials = service_account.Credentials.from_service_account_info(mapping_obj)
            if 'path_to_credentials' in kwargs:
                path = kwargs.pop('path_to_credentials')
                if path is not None:
                    credentials = service_account.Credentials.from_service_account_file(path)
            if 'credentials' in kwargs:
                kwargs.pop('credentials')
        self.client = storage.Client(credentials=credentials, **kwargs)

    def load(
        self,
        bucket_name: str,
        object_key: str,
        format: FileFormat = None,
        limit: int = QUERY_ROW_LIMIT,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from Google Cloud Storage into a Pandas data frame. This function will load at
        maximum 100,000 rows of data from the specified file.

        Args:
            import_config (Mapping, optional): Configuration settings for importing file from
            Google Cloud Storage. Defaults to None.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to 100000.
            read_config (Mapping, optional): Configuration settings for reading file into data
            frame. Defaults to None.

        Returns:
            DataFrame: The data frame constructed from the file in the Google Cloud Storage bucket.
        """
        if format is None:
            format = self._get_file_format(object_key)
        with self.printer.print_msg(
            f'Loading data frame from bucket \'{bucket_name}\' at key \'{object_key}\''
        ):
            bucket = self.client.get_bucket(bucket_name)
            blob = bucket.blob(object_key)
            if not blob.exists():
                return DataFrame()
            buffer = BytesIO(blob.download_as_string())
            return self._read(buffer, format, limit, **kwargs)

    def export(
        self, df: DataFrame, bucket_name: str, object_key: str, format: FileFormat = None, **kwargs
    ) -> None:
        """
        Exports data frame to a Google Cloud Storage bucket.

        Args:
            df (DataFrame): Data frame to export
            write_config (Mapping, optional): Configuration settings for writing data frame to
            specified format. Defaults to None.
            export_config (Mapping, optional): Configuration settings for exporting data frame
            to Google Cloud Storage. Defaults to None.
        """
        if format is None:
            format = self._get_file_format(object_key)

        with self.printer.print_msg(
            f'Exporting data frame to bucket \'{bucket_name}\' at key \'{object_key}\''
        ):
            bucket = self.client.get_bucket(bucket_name)
            blob = bucket.blob(object_key)
            buffer = BytesIO()
            self._write(df, format, buffer, **kwargs)
            buffer.seek(0)
            blob.upload_from_string(buffer.read())

    def exists(
        self, bucket_name: str, prefix: str
    ) -> bool:
        """
        Checks if content exists at a certain path in a Google Cloud Storage bucket.
        """
        bucket = self.client.get_bucket(bucket_name)
        blob = bucket.blob(prefix)
        return blob.exists()

    @classmethod
    def with_config(
        cls,
        config: BaseConfigLoader,
        **kwargs,
    ) -> 'GoogleCloudStorage':
        """
        Initializes Google Cloud Storage client from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        if ConfigKey.GOOGLE_SERVICE_ACC_KEY in config:
            kwargs['credentials_mapping'] = config[ConfigKey.GOOGLE_SERVICE_ACC_KEY]
        elif ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH in config:
            kwargs['path_to_credentials'] = config[ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH]
        else:
            raise ValueError(
                'No valid configuration settings found for Google Cloud Storage. You must specify '
                'either your service account key or the filepath to your service account key.'
            )
        return cls(**kwargs)



from mage_ai.shared.parsers import encode_complex
from mage_ai.shared.strings import camel_to_snake_case
import dateutil.parser
import json
import simplejson
import tornado.web
import traceback

META_KEY_LIMIT = '_limit'
META_KEY_OFFSET = '_offset'


class BaseHandler(tornado.web.RequestHandler):
    datetime_keys = []
    model_class = None

    def check_origin(self, origin):
        return True

    def get_bool_argument(self, name, default_value=None):
        value = self.get_argument(name, default_value)
        if type(value) is not str:
            return value
        return value.lower() in ('yes', 'true', 't', '1')

    def limit(self, results):
        limit = self.get_argument(META_KEY_LIMIT, None)
        offset = self.get_argument(META_KEY_OFFSET, None)
        if limit is not None:
            results = results.limit(limit)
        if offset is not None:
            results = results.offset(offset)
        return results

    def options(self, **kwargs):
        self.set_status(204)
        self.finish()

    def set_default_headers(self):
        methods = 'DELETE, GET, PATCH, POST, PUT, OPTIONS'
        self.set_header('Access-Control-Allow-Headers', '*')
        self.set_header('Access-Control-Allow-Methods', methods)
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Content-Type', 'application/json')

    def write(self, chunk):
        if type(chunk) is dict:
            chunk = simplejson.dumps(
                chunk,
                default=encode_complex,
                ignore_nan=True,
            )
        super().write(chunk)

    def write_error(self, status_code, **kwargs):
        if status_code == 500:
            self.set_status(200)
            exception = kwargs['exc_info'][1]
            self.write(
                dict(
                    error=dict(
                        code=status_code,
                        errors=traceback.format_stack(),
                        exception=str(exception),
                        message=traceback.format_exc(),
                    ),
                    url_parameters=self.path_kwargs,
                )
            )

    def write_model(self, model, **kwargs):
        key = camel_to_snake_case(self.model_class.__name__)
        self.write({key: model.to_dict(**kwargs)})

    def write_models(self, models):
        key = camel_to_snake_case(self.model_class.__name__) + 's'
        self.write({key: [m.to_dict() for m in models]})

    def get_payload(self):
        key = ''
        if self.model_class:
            key = camel_to_snake_case(self.model_class.__name__)

        payload = json.loads(self.request.body)
        if key != '':
            payload = payload.get(key, {})
        for key in self.datetime_keys:
            if payload.get(key) is not None:
                payload[key] = dateutil.parser.parse(payload[key])
        return payload


class BaseDetailHandler(BaseHandler):
    def get(self, model_id, **kwargs):
        model = self.model_class.query.get(int(model_id))
        self.write_model(model, **kwargs)

    def put(self, model_id, payload=None):
        model = self.model_class.query.get(int(model_id))
        payload = payload or self.get_payload()
        model.update(**payload)
        self.write_model(model)

    def delete(self, model_id):
        model = self.model_class.query.get(int(model_id))
        model.delete()
        self.write_model(model)



from .base import (
    META_KEY_LIMIT,
    BaseDetailHandler,
    BaseHandler,
)
from datetime import datetime
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.orchestration.db.models import (
    BlockRun,
    EventMatcher,
    PipelineRun,
    PipelineSchedule,
    pipeline_schedule_event_matcher_association_table,
)
from mage_ai.shared.hash import merge_dict
from sqlalchemy.orm import aliased, joinedload


class ApiBlockRunDetailHandler(BaseHandler):
    model_class = BlockRun

    def put(self, block_run_id):
        payload = self.get_payload()
        # Only allow update block run status
        status = payload.get('status')
        if status is not None:
            block_run = BlockRun.query.get(int(block_run_id))
            if status != block_run.status:
                block_run.update(status=status)
        self.write(dict(block_run=block_run.to_dict()))


class ApiAllBlockRunListHandler(BaseHandler):
    model_class = BlockRun

    def get(self):
        query = BlockRun.query

        a = aliased(BlockRun, name='a')
        b = aliased(PipelineRun, name='b')
        c = aliased(PipelineSchedule, name='c')

        columns = [
            a.block_uuid,
            a.completed_at,
            a.created_at,
            a.id,
            a.pipeline_run_id,
            a.status,
            a.updated_at,
            c.id.label('pipeline_schedule_id'),
            c.name.label('pipeline_schedule_name'),
        ]

        query = (
            BlockRun.
            select(*columns).
            join(b, a.pipeline_run_id == b.id).
            join(c, b.pipeline_schedule_id == c.id)
        )

        pipeline_run_id = self.get_argument('pipeline_run_id', None)
        if pipeline_run_id:
            query = (
                query.
                filter(a.pipeline_run_id == int(pipeline_run_id))
            )

        pipeline_uuid = self.get_argument('pipeline_uuid', None)
        if pipeline_uuid:
            query = (
                query.
                filter(c.pipeline_uuid == pipeline_uuid)
            )

        query = (
            query.
            order_by(a.created_at.desc(), a.completed_at.desc())
        ).all()
        collection = [r for r in query]

        self.write(dict(block_runs=collection))
        self.finish()


class ApiBlockRunListHandler(BaseHandler):
    model_class = BlockRun

    def get(self, pipeline_run_id):
        block_runs = BlockRun.query.filter(
            BlockRun.pipeline_run_id == int(pipeline_run_id),
        ).all()
        collection = [r.to_dict() for r in block_runs]

        self.write(dict(block_runs=collection))
        self.finish()


class ApiBlockRunLogHandler(BaseHandler):
    def get(self, block_run_id):
        block_run = BlockRun.query.get(int(block_run_id))
        self.write(
            dict(
                log=block_run.log_file.to_dict(include_content=True),
            ),
        )


class ApiBlockRunOutputHandler(BaseHandler):
    def get(self, block_run_id):
        block_run = BlockRun.query.get(int(block_run_id))
        outputs = block_run.get_outputs()
        self.write(dict(outputs=outputs))


def process_pipeline_runs(
    handler,
    pipeline_schedule_id=None,
    pipeline_uuid=None,
):
    results = (
        PipelineRun.
        query.
        options(joinedload(PipelineRun.block_runs)).
        options(joinedload(PipelineRun.pipeline_schedule))
    )
    if pipeline_schedule_id is not None:
        results = results.filter(PipelineRun.pipeline_schedule_id == pipeline_schedule_id)
    if pipeline_uuid is not None:
        results = results.filter(PipelineRun.pipeline_uuid == pipeline_uuid)            
    initial_results = \
        results.order_by(PipelineRun.execution_date.desc(), PipelineRun.id.desc())

    results = handler.limit(initial_results)

    collection = [r.to_dict(include_attributes=[
                                'block_runs',
                                'block_runs_count',
                                'pipeline_schedule_name',
                            ])
                  for r in results]

    # If runs from a certain execution date are included in the results, then
    # we want to include all of the attempts for that execution date. We need
    # to do this because the frontend groups retries.
    if handler.get_argument(META_KEY_LIMIT, None) is not None and \
        len(collection) > 0 and \
            (pipeline_uuid is not None or pipeline_schedule_id is not None):

        first_execution_date = results[0].execution_date
        additional_results = \
            initial_results.filter(PipelineRun.execution_date == first_execution_date)
        filter_dates = []
        if additional_results[0].id != results[0].id:
            filter_dates.append(collection[0]['execution_date'])

        last_execution_date = results[-1].execution_date
        additional_results = \
            initial_results.filter(PipelineRun.execution_date == last_execution_date)
        addons = [
            r.to_dict(include_attributes=[
                        'block_runs',
                        'block_runs_count',
                        'pipeline_schedule_name',
                    ])
            for r in additional_results
        ]
        filter_dates.append(collection[-1]['execution_date'])
        collection = \
            list(filter(lambda x: x['execution_date'] not in filter_dates, collection)) + addons

    handler.write(dict(pipeline_runs=collection, total_count=initial_results.count()))
    handler.finish()


class ApiAllPipelineRunListHandler(BaseHandler):
    datetime_keys = ['execution_date']
    model_class = PipelineRun

    def get(self):
        pipeline_uuid = self.get_argument('pipeline_uuid', None)
        process_pipeline_runs(self, pipeline_uuid=pipeline_uuid)


class ApiPipelineRunDetailHandler(BaseDetailHandler):
    model_class = PipelineRun
    
    def get(self, pipeline_run_id):
        pipeline_run = PipelineRun.query.get(int(pipeline_run_id))
        block_runs = pipeline_run.block_runs

        pipeline_run_dict = pipeline_run.to_dict()
        block_runs_json = []
        for r in block_runs:
            block_run = r.to_dict()
            block_run['pipeline_schedule_id'] = pipeline_run.pipeline_schedule_id
            block_run['pipeline_schedule_name'] = pipeline_run.pipeline_schedule.name
            block_runs_json.append(block_run)
        pipeline_run_dict['block_runs'] = block_runs_json

        self.write(dict(pipeline_run=pipeline_run_dict))

    def put(self, pipeline_run_id):
        """
        Only allow cancelling a pipeline run with this endpoint
        """
        payload = self.get_payload()
        pipeline_run = PipelineRun.query.get(int(pipeline_run_id))
        if payload.get('status') == PipelineRun.PipelineRunStatus.CANCELLED:
            from mage_ai.orchestration.pipeline_scheduler import PipelineScheduler
            PipelineScheduler(pipeline_run).stop()
        self.write(dict(pipeline_run=pipeline_run.to_dict()))


class ApiPipelineRunListHandler(BaseHandler):
    datetime_keys = ['execution_date']
    model_class = PipelineRun

    def get(self, pipeline_schedule_id):
        process_pipeline_runs(self, pipeline_schedule_id=int(pipeline_schedule_id))

    def post(self, pipeline_schedule_id):
        pipeline_schedule = PipelineSchedule.query.get(int(pipeline_schedule_id))

        payload = self.get_payload()
        payload['pipeline_schedule_id'] = pipeline_schedule.id
        payload['pipeline_uuid'] = pipeline_schedule.pipeline_uuid
        if payload.get('execution_date') is None:
            payload['execution_date'] = datetime.now()
        pipeline_run = PipelineRun.create(**payload)

        from mage_ai.orchestration.pipeline_scheduler import PipelineScheduler
        PipelineScheduler(pipeline_run).start(should_schedule=False)

        self.write(dict(pipeline_run=pipeline_run.to_dict()))


class ApiPipelineRunLogHandler(BaseHandler):
    def get(self, pipeline_run_id):
        pipeline_run = PipelineRun.query.get(int(pipeline_run_id))
        self.write(
            dict(
                log=pipeline_run.log_file.to_dict(include_content=True),
            ),
        )


class ApiPipelineScheduleDetailHandler(BaseDetailHandler):
    datetime_keys = ['start_time']
    model_class = PipelineSchedule

    def get(self, pipeline_schedule_id):
        include_attributes = []
        if self.get_bool_argument('include_event_matchers', False):
            include_attributes.append('event_matchers')
        super().get(pipeline_schedule_id, include_attributes=include_attributes)

    def put(self, pipeline_schedule_id):
        pipeline_schedule = PipelineSchedule.query.get(int(pipeline_schedule_id))
        payload = self.get_payload()

        arr = payload.pop('event_matchers', None)
        event_matchers = []
        if arr is not None:
            if len(arr) >= 1:
                event_matchers = EventMatcher.upsert_batch(
                    [merge_dict(p, dict(pipeline_schedule_ids=[pipeline_schedule_id]))
                     for p in arr],
                )

            ems = (
                EventMatcher.
                query.
                join(
                    pipeline_schedule_event_matcher_association_table,
                    EventMatcher.id ==
                    pipeline_schedule_event_matcher_association_table.c.event_matcher_id
                ).
                join(
                    PipelineSchedule,
                    PipelineSchedule.id ==
                    pipeline_schedule_event_matcher_association_table.c.pipeline_schedule_id
                ).
                filter(
                    PipelineSchedule.id == int(pipeline_schedule_id),
                    EventMatcher.id.not_in([em.id for em in event_matchers]),
                )
            )
            for em in ems:
                new_ids = [schedule for schedule in em.pipeline_schedules
                           if schedule.id != int(pipeline_schedule_id)]
                ps = [p for p in PipelineSchedule.query.filter(PipelineSchedule.id.in_(new_ids))]
                em.update(pipeline_schedules=ps)

        pipeline_schedule.update(**payload)

        extra_data = {}
        include_attributes = []

        if len(event_matchers) >= 1:
            extra_data['event_matchers'] = [em.to_dict() for em in event_matchers]
        else:
            include_attributes.append('event_matchers')

        pipeline_schedule_data = merge_dict(
            pipeline_schedule.to_dict(include_attributes=include_attributes),
            extra_data,
        )

        self.write(dict(pipeline_schedule=pipeline_schedule_data))

    def delete(self, pipeline_schedule_id):
        super().delete(pipeline_schedule_id)


class ApiPipelineScheduleListHandler(BaseHandler):
    datetime_keys = ['start_time']
    model_class = PipelineSchedule

    def get(self, pipeline_uuid=None):
        try:
            if pipeline_uuid is None:
                results = PipelineSchedule.query.all()
                collection = [s.to_dict() for s in results]
            else:
                pipeline = Pipeline.get(pipeline_uuid)
                results = (
                    PipelineSchedule.
                    query.
                    options(joinedload(PipelineSchedule.event_matchers)).
                    options(joinedload(PipelineSchedule.pipeline_runs)).
                    filter(PipelineSchedule.pipeline_uuid == pipeline.uuid).
                    order_by(PipelineSchedule.start_time.desc(), PipelineSchedule.id.desc())
                )
                results = self.limit(results)
                collection = [r.to_dict(include_attributes=['event_matchers', 'pipeline_runs_count'])
                              for r in results]
                collection.sort(key=lambda d: d['id'], reverse=True)
        except Exception as err:
            raise err
            collection = []

        self.write(dict(pipeline_schedules=collection))
        self.finish()

    def post(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)

        payload = self.get_payload()
        payload['pipeline_uuid'] = pipeline.uuid
        pipeline_schedule = PipelineSchedule.create(**payload)

        self.write(dict(pipeline_schedule=pipeline_schedule.to_dict()))



import os

SERVER_HOST = os.getenv('HOST', 'localhost')
SERVER_PORT = os.getenv('PORT', 5789)

DATA_PREP_SERVER_PORT = os.getenv('DATA_PREP_SERVER_PORT', 6789)

DATAFRAME_OUTPUT_SAMPLE_COUNT = 10

VERSION = '0.4.19'



from email.message import EmailMessage
from mage_ai.services.email.config import EmailConfig
import smtplib


def send_email(config: EmailConfig, subject: str = None, message: str = None):
    if subject is None or message is None:
        return
    msg = EmailMessage()
    msg['Subject'] = subject
    msg['From'] = config.smtp_mail_from
    msg['To'] = ', '.join(config.to_emails)
    msg.set_content(message)
    print(msg)
    server = smtplib.SMTP(config.smtp_host, config.smtp_port)
    server.starttls()
    server.login(config.smtp_user, config.smtp_password) 
    server.send_message(msg)
    server.quit()
    print('Successfully sent the mail.')



from mage_ai.data_preparation.logger_manager import LoggerManager
from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.shared.hash import merge_dict
from typing import Callable, Dict
import json
import requests


class BlockExecutor:
    def __init__(self, pipeline, block_uuid, execution_partition=None):
        self.pipeline = pipeline
        self.block_uuid = block_uuid
        self.block = self.pipeline.get_block(block_uuid)
        self.execution_partition = execution_partition
        logger_manager = LoggerManager.get_logger(
            pipeline_uuid=self.pipeline.uuid,
            block_uuid=self.block_uuid,
            partition=self.execution_partition,
        )
        self.logger = DictLogger(logger_manager)

    def execute(
        self,
        analyze_outputs: bool = False,
        callback_url: str = None,
        global_vars: Dict = None,
        update_status: bool = False,
        on_complete: Callable[[str], None] = None,
        on_failure: Callable[[str], None] = None,
        on_start: Callable[[str], None] = None,
        **kwargs,
    ) -> None:
        tags = self._build_tags(**kwargs.get('tags', {}))

        self.logger.info(f'Start executing block with {self.__class__.__name__}.', **tags)
        if on_start is not None:
            on_start(self.block_uuid)
        try:
            self._execute(
                analyze_outputs=analyze_outputs,
                callback_url=callback_url,
                global_vars=global_vars,
                update_status=update_status,
                **kwargs,
            )
        except Exception as e:
            self.logger.info('Failed to execute block.', **tags)
            if on_failure is not None:
                on_failure(self.block_uuid)
            elif callback_url is not None:
                self.__update_block_run_status(callback_url, 'failed')
            raise e
        self.logger.info(f'Finish executing block with {self.__class__.__name__}.', **tags)
        if on_complete is not None:
            on_complete(self.block_uuid)
        elif callback_url is not None:
            self.__update_block_run_status(callback_url, 'completed', tags)

    def _execute(
        self,
        analyze_outputs: bool = False,
        callback_url: str = None,
        global_vars: Dict = None,
        update_status: bool = False,
        **kwargs,
    ):
        self.block.execute_sync(
            analyze_outputs=analyze_outputs,
            execution_partition=self.execution_partition,
            global_vars=global_vars,
            logger=self.logger,
            run_all_blocks=True,
            update_status=update_status,
        )
        self.block.run_tests(
            logger=self.logger,
            update_tests=False,
        )

    def __update_block_run_status(self, callback_url: str, status: str, tags: dict):
        response = requests.put(
            callback_url,
            data=json.dumps({
                'block_run': {
                    'status': status,
                },
            }),
            headers={
                'Content-Type': 'application/json',
            },
        )
        self.logger.info(
            f'Callback response: {response.text}',
            **tags,
        )

    def _build_tags(self, **kwargs):
        return merge_dict(kwargs, dict(
            block_type=self.block.type,
            block_uuid=self.block.uuid,
            pipeline_uuid=self.pipeline.uuid,
        ))



from contextlib import redirect_stdout
from mage_ai.data_preparation.executors.block_executor import BlockExecutor
from mage_ai.data_preparation.logger_manager import StreamToLogger
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.templates.utils import template_env
from mage_ai.services.aws.emr import emr
from mage_ai.services.aws.emr.resource_manager import EmrResourceManager
from mage_ai.services.aws.s3 import s3
from mage_ai.shared.hash import merge_dict
from typing import Dict
import os


class PySparkBlockExecutor(BlockExecutor):
    def __init__(self, pipeline: Pipeline, block_uuid: str, execution_partition: str = None):
        super().__init__(pipeline, block_uuid, execution_partition=execution_partition)
        self.resource_manager = EmrResourceManager(
            pipeline.repo_config.s3_bucket,
            pipeline.repo_config.s3_path_prefix,
        )
        self.s3_bucket = pipeline.repo_config.s3_bucket
        self.s3_path_prefix = pipeline.repo_config.s3_path_prefix
        self.executor_config = self.pipeline.repo_config.emr_config
        if self.block.executor_config is not None:
            self.executor_config = merge_dict(self.executor_config, self.block.executor_config)

    def _execute(
        self,
        analyze_outputs: bool = False,
        global_vars: Dict = None,
        update_status: bool = False,
        **kwargs,
    ) -> None:
        """
        Run block in a spark cluster
        1. Upload block execution script to S3
        2. Launch or connect to an EMR spark cluster
        3. Submit a spark job
        """
        stdout = StreamToLogger(self.logger)
        with redirect_stdout(stdout):
            self.upload_block_execution_script(global_vars=global_vars)
            self.resource_manager.upload_bootstrap_script()
            self.submit_spark_job()

    @property
    def spark_script_path(self) -> str:
        return os.path.join('s3://', self.s3_bucket, self.spark_script_path_key)

    @property
    def spark_script_path_key(self) -> str:
        return os.path.join(
            self.s3_path_prefix,
            f'scripts/{self.pipeline.uuid}/{self.block_uuid}.py',
        )

    def upload_block_execution_script(self, global_vars: Dict = None) -> None:
        execution_script_code = template_env.get_template(
            'pipeline_execution/spark_script.jinja',
        ).render(
            block_uuid=f'\'{self.block_uuid}\'',
            execution_partition_str=f'\'{self.execution_partition}\''
                                    if self.execution_partition is not None else None,
            global_vars=global_vars,
            pipeline_config=self.pipeline.to_dict(include_content=True),
            pipeline_uuid=self.pipeline.uuid,
            repo_config=self.pipeline.repo_config.to_dict(remote=True),
            spark_log_path=self.resource_manager.log_uri,
        )

        s3.Client(self.s3_bucket).upload(self.spark_script_path_key, execution_script_code)

    def submit_spark_job(self):
        step = {
            'name': f'run_mage_block_{self.pipeline.uuid}_{self.block_uuid}',
            'script_uri': self.spark_script_path,
            'script_args': [],
        }
        return emr.submit_spark_job(
            cluster_name=step['name'],
            steps=[step],
            bootstrap_script_path=self.resource_manager.bootstrap_script_path,
            emr_config=self.executor_config,
            log_uri=self.resource_manager.log_uri,
        )
